/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2010   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * fls_chanvese.cc
 *
 * A simple console application intended for image segmentation based on the minimization of the Chan-Vese model using the Nilsson-Heyden algorithm
 *
 * Martin Maka (xmaska@fi.muni.cz) 2006 - 2011
 * 
 */

#include <iostream>
#include <time.h>
#include "image3d.h"
#include "FastLevelSet.h"

/////////////////////////////////////////////////////////////////
//                                                             //
//                         Local Functions                     //
//                                                             //
/////////////////////////////////////////////////////////////////

/**********************************************************************/

/** Load image data from given file. Return false if any error occurs. */
template <class T> bool LoadImage(i3d::Image3d<T> &img, const char *fname)
{
	try 
	{
		img.ReadImage(fname);
	}
	catch (i3d::IOException &e)
	{
		printf("i3d::IOException: %s\n", e.what.c_str());
		return false;
	}
	catch (i3d::InternalException &e)
	{
		printf("i3d::InternalException: %s\n", e.what.c_str());
		return false;
	}
	catch (std::bad_alloc &)
	{
		printf("Not enough memory\n");
		return false;
	}        
	catch (...)
	{
		printf("Unknown exception caught\n");
		return false;
	}

	return true;
}

/**********************************************************************/

/** Load an image from given file. Return false if any error occurs. */
template <class T> bool ReadImage(i3d::Image3d<T> &img, const char *fname)
{
	try 
	{
		i3d::ImgVoxelType vt = i3d::ReadImageType(fname);

		if (vt == i3d::Gray16Voxel)
		{
			return LoadImage(img, fname);
		}
		else if (vt == i3d::BinaryVoxel)
		{
			return LoadImage(img, fname);
		}
		else
		{
			printf("Unsupported image type (%s)\n", i3d::VoxelTypeToString(vt).c_str());
			return false;
		}
	}
	catch (i3d::IOException &e)
	{
		printf("i3d::IOException: %s\n", e.what.c_str());
		return false;
	}
}

/**********************************************************************/

/////////////////////////////////////////////////////////////////
//                                                             //
//                           Entry Point                       //
//                                                             //
/////////////////////////////////////////////////////////////////

/**********************************************************************/

/** Entry point of the console application. */
int main(int argc, char **argv)
{
	if (argc != 9)
	{
		 std::cout << "A simple console application intended for image segmentation based on the minimization of the Chan-Vese model using the Nilsson-Heyden algorithm \n\n"
	                  "Copyright (C) 2006 - 2011 Centre for Biomedical Image Analysis \n"
					  "This is free software; see the source for copying conditions. There is NO\nwarranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n";
    
		 std::cout << "Usage: "<< argv[0] << " input_image initial_interface_image output_image preserve_topology curvature_impact foreground_impact background_impact num_iter\n";

		 return -1;
	}

	i3d::Image3d<i3d::GRAY16> input;
	i3d::Image3d<bool> initial_interface, output;
	i3d::Image3d<i3d::fls::Type::Prec> speed;

	bool preserve_topology = (atoi(argv[4]) == 1);
	i3d::fls::Type::Prec curvature_impact = i3d::fls::Type::Prec(atof(argv[5]));
	i3d::fls::Type::Prec foreground_impact = i3d::fls::Type::Prec(atof(argv[6]));
	i3d::fls::Type::Prec background_impact = i3d::fls::Type::Prec(atof(argv[7]));
	i3d::fls::Type::Size num_iter = i3d::fls::Type::Size(atoi(argv[8]));

	// load an input image
	if (!ReadImage(input, argv[1]))
	{
		std::cout << "The input image '" << argv[1] << "' cannot be opened!" << std::endl;
		return -1;
	}

	// load an image containing an initial interface
	if (!ReadImage(initial_interface, argv[2]))
	{
		std::cout << "The interface image '" << argv[2] << "' cannot be opened!" << std::endl;
		return -1;
	}

	try
	{
		// store a timestamp at which the computation begins
		time_t start = clock();

		// create a speed image intended for the Chan-Vese model
		i3d::fls::RegionSpeedImage(input, speed);

		// create a routine intended for the minimization of the Chan-Vese model
		i3d::fls::Routine<int> *routine = i3d::fls::GlobChanVeseRoutine(speed, initial_interface, NULL, preserve_topology, 3, curvature_impact, foreground_impact, background_impact);

		// initialize the routine
		routine -> Initialize();

		// perform given number of iterations
		routine -> Iterate(num_iter);

		// copy the result to the output image
		output.CopyMetaData(input);
		routine -> Mask(output.begin(), i3d::fls::ResultCreator::InteriorAndInterface);

		// delete the routine
		delete routine;

		// display the total execution time
		std::cout << "Execution time: " << float(clock() - start) /  float(CLOCKS_PER_SEC) << "s.\n\n";

		// save the output image
		output.SaveImage(argv[3]);
	}
	catch (i3d::InternalException &e)
	{
		std::cout << "i3d::InternalException: " << e.what.c_str() << std::endl;
		return -1;
	}
	catch (i3d::IOException &e)
	{
		std::cout << "i3d::IOException: " << e.what.c_str() << std::endl;
		return -1;
	}
	catch (std::bad_alloc &)
	{
		std::cout << "Not enough memory\n";
		return -1;
	}
	catch (...)
	{
		std::cout <<  "Unknown exception caught\n";
		return -1;
	}
		
	return 0;
}
