
/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * FILE: imgPNG.cc
 *
 * I/O routines for PNG images
 *
 * Maros Gonda (mano@mail.muni.cz) 2011
 *
 */

#include <string.h>
#include "i3d_config.h"

#ifdef WITH_PNG

#include <string>
#include <string.h>

#include "imgPNG.h"
#include "zlib.h"

extern "C" {
#include <pngconf.h>
}

using namespace std;

namespace i3d {



/************************************************************/
/*                        PNGReader                         */
/************************************************************/
PNGReader::PNGReader(const char *fname, const VOI<PIXELS> *voi):ImageReader(fname, voi) {
	
	
	// open file and load PNG header of file
	 
	png_byte signature[8];
	
	if ((fd = fopen(fname, "rb")) == NULL)
		throw IOException("PNG: Unable to open "+ std::string(fname) + ".");
	if (fread(&signature, 1, 8, fd) != 8)
		throw IOException("PNG: Unable to read PNG signature.");
	
	
	// check loaded PNG header of file
	if (png_sig_cmp(signature, 0, 8))
		throw IOException("PNG: Set file is not PNG format.");
	
	
	// create image structure and image info structure
	png_image_struct = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_image_struct == NULL) {
		fclose(fd);
		throw IOException("PNG: Unable to create PNG structure.");
	}
	png_image_info_struct = png_create_info_struct(png_image_struct);
	if (png_image_info_struct == NULL) {
		fclose(fd);
		png_destroy_read_struct(&png_image_struct, NULL, NULL);
		throw IOException("PNG: Unable to create PNG info structure.");
	}
	
	
	// set error handling with the file
	if (setjmp(png_jmpbuf(png_image_struct))) {
		fclose(fd);
		png_destroy_read_struct(&png_image_struct, &png_image_info_struct, NULL);
		throw IOException("PNG: Unable to read PNG file");
	}
	
}





/***************************************************************************/
/*                          LoadImageInfo                                  */
/***************************************************************************/
void PNGReader::LoadImageInfo() {
	
	
	int bit_depth, color_type, interlace_type, unit=PNG_RESOLUTION_UNKNOWN;
	png_uint_32 res_x=0, res_y=0, width, height;
	
	
	// initialise IO operations for PNG files and set image reading for info beginning
	png_init_io(png_image_struct, fd);
	png_set_sig_bytes(png_image_struct, 8);
	
	
	// load information about file and main 
	png_read_info(png_image_struct, png_image_info_struct);

	png_get_IHDR(png_image_struct, 
					 png_image_info_struct, 
					 &width, 
					 &height, 
					 &bit_depth, 
					 &color_type, 
					 &interlace_type, NULL, NULL);

	header.size.x = width;
	header.size.y = height;
	
	
	#ifdef I3D_DEBUG
	cout << "PNG: Loading image color mode" << endl;
	#endif
	
	switch (color_type) {
		case (PNG_COLOR_TYPE_GRAY): 
		case (PNG_COLOR_TYPE_GRAY_ALPHA): {
				if (bit_depth==1) {
					header.itype = BinaryVoxel;
					png_set_expand_gray_1_2_4_to_8(png_image_struct);
				} else if (bit_depth==16)
					header.itype = Gray16Voxel;
				else {
					header.itype = Gray8Voxel;
					if (bit_depth<8) png_set_expand_gray_1_2_4_to_8(png_image_struct);
				}
				png_set_strip_alpha(png_image_struct);
				break;
		}
		case (PNG_COLOR_MASK_PALETTE):
		case (PNG_COLOR_TYPE_PALETTE): {
				png_set_palette_to_rgb(png_image_struct);
				
				// unpack bits into 8 bits color depth
				png_set_packing(png_image_struct);
				header.itype = RGBVoxel;
				break;
		}
		case (PNG_COLOR_TYPE_RGB_ALPHA): {
				// (possible bit depths 8, 16)
				// remove alpha channel
				png_set_strip_alpha(png_image_struct);
		}
		case (PNG_COLOR_TYPE_RGB): {
				// (possible bit depths 8, 16)
				if (bit_depth==16)
					header.itype = RGB16Voxel;
				else //if (bit_depth==8)
					header.itype = RGBVoxel;
				break;
		}
		
		default:
			
			header.itype = UnknownVoxel;
	}
	
	
	
	
	
	#ifdef I3D_DEBUG
	cout << "PNG: Loading image background" << endl;
	#endif
	
	// set default color of background for transparent images
	png_color_16 background;
	png_color_16p backgroundp = &background;
	
	
	if (png_get_bKGD(png_image_struct, png_image_info_struct, &backgroundp)) {
		png_set_background(png_image_struct, backgroundp, PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
	} else {
		// set background color as black 
		background.red   = 0;
		background.green = 0;
		background.blue  = 0;
		background.gray  = 0;
		background.index = 0;
		
		png_set_background(png_image_struct, backgroundp, PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);
	}
	
	
	
	
	// load image resolution
	png_get_pHYs(png_image_struct, png_image_info_struct, &res_x, &res_y, &unit);
	if (unit==PNG_RESOLUTION_METER && res_x && res_y) {
		header.resolution = new Vector3d<float>(static_cast<float>(res_x), static_cast<float>(res_y), 0.0);		
	}
	
	
	// load offset info
	header.offset.x = (float) png_get_x_offset_microns(png_image_struct, png_image_info_struct);
	header.offset.y = (float) png_get_y_offset_microns(png_image_struct, png_image_info_struct);
	
	
	// load text image description
	LoadDescription();
	
}


/****************************************************************************
		LoadDescription
	loads image text information
 ****************************************************************************/
void PNGReader::LoadDescription() {
	
		png_uint_32 comments,i;
		png_textp description;
		
		// load image description from file
		comments = png_get_text(png_image_struct, png_image_info_struct, &description, NULL);
		
		
#			ifdef I3D_DEBUG
			cout << "-----------------------------" << endl
			     << "  Loading image description:  " << endl;
#			endif
		
		for (i = 0; i <comments; i++)
			header.description[string(description[i].key)] = string(description[i].text);
		
		
		
#			ifdef I3D_DEBUG
			
			map<string, string>::iterator iter;
			for (iter=header.description.begin(); iter!=header.description.end(); iter++)
				cout << "  [" << iter->first << "]: " << iter->second << endl;
			cout << "-----------------------------" << endl;
#			endif
	
	
}





/***************************************************************************/
/* ~PNGReader */
/***************************************************************************/
PNGReader::~PNGReader() {
	png_destroy_read_struct(&png_image_struct, &png_image_info_struct, NULL);
	fclose(fd);
}






/***************************************************************************
   LoadData 
   - Reads image data and crops it if set
 ***************************************************************************/ 
template <class T> void PNGReader::LoadData(T *data) {
	
	png_uint_32 i, offsetx=0, offsetY = 0;
	
	png_read_update_info(png_image_struct, png_image_info_struct);
	if (png_image_struct == NULL)
	    return;
	
	// invalid: png_bytep rows[header.size.y];
	png_bytep* rows = (png_bytep*) malloc(header.size.y * sizeof(png_bytep));

	for (i=0; i<header.size.y; i++) {
		rows[i] = static_cast<png_bytep>(png_malloc(png_image_struct, png_get_rowbytes(png_image_struct, png_image_info_struct)));
	}
	
	png_read_image(png_image_struct, rows);
	
	
	
	if (crop) { 
		if (crop->size.z >0 && crop->size.x>0 && crop->size.y >0) {
			
			if ((header.size.x >= crop->size.x + crop->offset.x) && (header.size.y >= crop->size.y + crop->offset.y)) {
				header.size = crop->size;
				offsetY = crop->offset.y;
				offsetx = crop->offset.x;
				
#				ifdef I3D_DEBUG
				cout << "Cropping image to: [" << offsetx << "-" << (header.size.x+offsetx) << ", " << offsetY << "-" << (header.size.y+offsetY) << "]" << endl;
#				endif
				
			}
		} else {
			throw IOException("PNG: Can't load cropped image with zero dimension");
		}
	}
#	ifdef I3D_DEBUG
		cout << "Loading image coordinates: [" << offsetx << "-" << (header.size.x+offsetx) << ", " << offsetY << "-" << (header.size.y+offsetY) << "]" << endl;
#	endif	
	
	for (i=offsetY; i < header.size.y+offsetY; i++) {
			memcpy(data, rows[i]+offsetx*sizeof(T),sizeof(T)*header.size.x);
			data = data + header.size.x;
			
	}
	
	
	
	png_read_end(png_image_struct, png_image_info_struct);
	for (i=0; i<header.size.y; i++)
		png_free(png_image_struct, rows[i]);
	free(rows);
}








/***************************************************************************/
/* SaveImageInfo */
/***************************************************************************/
void PNGWriter::SaveImageInfo() {
	
	int bit_depth, color_type;
	
	
	// choose PNG color type and bit depth according to itype of image
	switch (header.itype) {
		case (BinaryVoxel): {
				color_type = PNG_COLOR_TYPE_GRAY;
				bit_depth = 1;
				break;
		}
		case (Gray8Voxel): {
				color_type = PNG_COLOR_TYPE_GRAY;
				bit_depth = 8;
				break;
		}
		case (Gray16Voxel): {
				color_type = PNG_COLOR_TYPE_GRAY;
				bit_depth = 16;
				break;
		}
		case (RGBVoxel): {
				color_type = PNG_COLOR_TYPE_RGB;
				bit_depth = 8;
				break;
		}
		case (RGB16Voxel): {
				color_type = PNG_COLOR_TYPE_RGB;
				bit_depth = 16;
				break;
		}
		default: {
			throw IOException("PNG: Unknown Voxel type.");
			break;
		}
	}
	
	
	// set PNG Header
	png_set_IHDR(png_image_struct, png_image_info_struct, 
			header.size.x, header.size.y, bit_depth, color_type,
			PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, 
			PNG_FILTER_TYPE_DEFAULT);
	
	
	
	// if picture offset and resolution were set, write them info into file
	if (header.offset.x && header.offset.y)
		png_set_oFFs(png_image_struct, png_image_info_struct, static_cast<int>(header.offset.x), static_cast<int>(header.offset.y), PNG_OFFSET_MICROMETER);
	
	if (header.resolution)
		png_set_pHYs(png_image_struct, png_image_info_struct, static_cast<int>(header.resolution->x), static_cast<int>(header.resolution->y), PNG_RESOLUTION_METER);
	
	
	SaveDescription();
	
	
	png_write_info(png_image_struct, png_image_info_struct);
	
}




/****************************************************************************
	Saves image description into info struct
 ****************************************************************************/
void PNGWriter::SaveDescription() {
		
		png_text description[10];
		
		
		// key text can be in range 1-79 characters
		// PNG specification keywords:
		// 	Title		Author
		// 	Description	Copyright
		// 	Creation Time	Software
		// 	Disclaimer	Warning
		// 	Source		Comment
		// 
		
#		ifdef I3D_DEBUG
			cout << "---------------------------------" << endl
			     << "  PNG: Writing image description:" << endl;
#		endif
			
		map<string, string>::iterator iter;
		int i=0;
		for (iter=header.description.begin(); iter!=header.description.end(); iter++) {
			
			description[i].key = const_cast<char*>(iter->first.c_str());
			description[i].text = const_cast<char*>(iter->second.c_str());
			
			// Compresion is recommended only for 1k of chars
			description[i].compression = (header.compression ? PNG_TEXT_COMPRESSION_zTXt : PNG_TEXT_COMPRESSION_NONE);
			
#			ifdef I3D_DEBUG
				cout << "[" << description[i].key << "]: " << description[i].text << endl;
#			endif
			i++;
		}
		
#		ifdef I3D_DEBUG
			cout << "---------------------------------" << endl;
#		endif
		// If libPNG supports UTF-8 with iTXt chunks add
		// 	description[0].lang = "en";
		// 	if (header.compression)
		// 		description[0].compression = PNG_ITXT_COMPRESSION_zTXt;
		// 	else
		// 		description[0].compression = PNG_ITXT_COMPRESSION_NONE;
		
		
		

			
		
		// write image description 
		png_set_text(png_image_struct, png_image_info_struct, description, header.description.size());
		
		
	
}






/*************************************************************************
	SaveData 
   - save one slice into file
************************************************************************/
template <class T> void PNGWriter::SaveData(const T *data) {
	
	png_uint_32 i;
	png_bytep byt = (png_bytep)data;
	png_bytep row;
	
	
	
	// take care of correct data saving according to specified bit depth
	png_set_packing(png_image_struct);
	
	
	
	// copy data into one row and save it in file
	row = (png_bytep)malloc(header.size.x*sizeof(T));
	
	for (i=0; i<header.size.y; i++) {
		memcpy(row, byt+i*header.size.x*sizeof(T), header.size.x*sizeof(T));
		png_write_row(png_image_struct, row);
	}
	
	
	// finnish writing and clean memory
	free(row);
	png_write_end(png_image_struct, png_image_info_struct);
	
}





/****************************************************************************
	PNGWriter
 ****************************************************************************/
PNGWriter::PNGWriter(const char *fname) : ImageWriter(fname) {
	
	if ((fd = fopen(fname, "wb")) == NULL)
		throw IOException("PNG: Unable to create "+ std::string(fname) + ". ");
	
	
	// create basic image structure
	png_image_struct = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_image_struct == NULL)
		throw IOException("PNG: Unable to create image structure.");
	
	
	// create basic image info structure
	png_image_info_struct = png_create_info_struct(png_image_struct);
	if (png_image_info_struct == NULL) {
		png_destroy_write_struct(&png_image_struct, NULL);
		throw IOException("PNG: Unable to create image info structure.");
	}
	
	
	// inicialise IO operations
	png_init_io(png_image_struct, fd);
	
	
	
	// set compression level - full or without compression
	if (header.compression)
		png_set_compression_level(png_image_struct, Z_BEST_COMPRESSION);
	else 
		png_set_compression_level(png_image_struct, Z_NO_COMPRESSION);
}




/***************************************************************************/
/* ~PNGWriter */
/***************************************************************************/
PNGWriter::~PNGWriter() {
	png_destroy_write_struct(&png_image_struct, &png_image_info_struct);
	fclose(fd);
}


}
#endif
