/*
* i3dlib - image manipulation library
*
* Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Library General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Library General Public License for more details.
*
* You should have received a copy of the GNU Library General Public
* License along with this library; if not, write to the Free
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/** \file component_tree.cc
\brief Part of mathematical morphology routines - implementation.

This file contains bodies of mathematical morphology routines dealing with 
the component trees.
*/ 

#include <typeinfo>
#include "component_tree.h"
#include "i3dassert.h"

#ifdef __GNUG__
#	pragma implementation
#endif

namespace i3d
{
	//--------------------------------------------------------------------------
	// Variation on the Tarjan's union-fing algorithm
	//--------------------------------------------------------------------------
	/** Representation of the disjoint sets insired by:
	R. E. Tarjan, Efficiency of a good but not linear set union algorithm, J. ACM, vol. 22, pp. 215225, 1975.
	*/
	class CCompTreeUnionFind
	{
	public:
		/// initializes union find class for storage of specified number of elements (initialy disjoint sets)
		CCompTreeUnionFind(size_t numOfElements)
		{
			parents = new size_t[numOfElements];
			for (size_t i = 0; i < numOfElements; ++i)
				parents[i] = i;
			ranks = new size_t[numOfElements];
			memset(ranks, 0, numOfElements * sizeof(size_t));
		}
		/// frees all the allocated resources
		~CCompTreeUnionFind()
		{
			delete[] parents;
			delete[] ranks;
		}

		/// returs representative of the set where the given element is stored
		size_t Find(size_t element)
		{
			// use the path compression concept - update the parents[element] for each element on the path to the representative
			if (parents[element] != element)
				parents[element] = Find(parents[element]);
			return parents[element];
		}
		/// joins two sets together, should be called on the set representatives, returns the new common representative
		size_t Link(size_t element1, size_t element2)
		{
			if (ranks[element1] > ranks[element2])
				std::swap(element1, element2);
			if (ranks[element1] == ranks[element2])
				++ranks[element2];
			parents[element1] = element2;
			return element2;
		}

	private:
		/// parents of the elementes
		size_t *parents;
		/// ranks of the elementes
		size_t *ranks;
	};

	//--------------------------------------------------------------------------
	// Component three node list implementation
	//--------------------------------------------------------------------------

	template <typename T>
	I3D_ALGO_EXPORT SCompTreeNodeListElem<T>::SCompTreeNodeListElem(SCompTreeNode<T> *value)
	{
		this->value = value;
		next = NULL;
	}

	template <typename T>
	I3D_ALGO_EXPORT SCompTreeNodeList<T>::SCompTreeNodeList()
	{
		first = last = NULL;
	}

	/// destructor that frees all the allocated resources
	template <typename T>
	I3D_ALGO_EXPORT SCompTreeNodeList<T>::~SCompTreeNodeList()
	{
		while (first)
		{
			SCompTreeNodeListElem<T> *elemToDelete = first;
			first = elemToDelete->next;
			delete elemToDelete;
		}
		first = last = NULL;
	}

	/// connects given second list at the end this list, second list is empty at the end
	template <typename T>
	I3D_ALGO_EXPORT void SCompTreeNodeList<T>::Merge(SCompTreeNodeList<T> &second)
	{
		if (!second.first)
			return;
		if (last)
			last->next = second.first;
		else
			first = second.first;
		last = second.last;
		second.first = second.last = NULL;
	}

	/// inserts and element at the end of this list
	template <typename T>
	I3D_ALGO_EXPORT void SCompTreeNodeList<T>::PushBack(SCompTreeNode<T> *value)
	{
		SCompTreeNodeListElem<T> *newCompTreeNodeListElem = new SCompTreeNodeListElem<T>(value);
		if (last)
			last->next = newCompTreeNodeListElem;
		else
			first = newCompTreeNodeListElem;
		last = newCompTreeNodeListElem;
	}

	//--------------------------------------------------------------------------
	// Component three node methods
	//--------------------------------------------------------------------------

	template <typename T>
	I3D_ALGO_EXPORT SCompTreeNode<T>::SCompTreeNode(const T &level)
	{
		this->level = level;
		attributes = NULL;
	}

	template <typename T>
	I3D_ALGO_EXPORT SCompTreeNode<T>::~SCompTreeNode()
	{
		for (SCompTreeNodeListElem<T> *i = children.first; i != NULL; i = i->next)
			delete i->value;
	}

	template I3D_ALGO_EXPORT struct SCompTreeNode<i3d::GRAY8>;
	template I3D_ALGO_EXPORT struct SCompTreeNode<i3d::GRAY16>;

	//--------------------------------------------------------------------------
	// Component three node (parent, height, area, volume) attributes handler methods
	//--------------------------------------------------------------------------

	template <typename T>
	I3D_ALGO_EXPORT CCompTreeNodeAttribHandlerPHAV<T>::CCompTreeNodeAttribHandlerPHAV()
	{
		building = true;
	}

	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::NodeCreated(SCompTreeNode<T> *node, const T &level)
	{
		node->attributes = malloc(sizeof(T) + sizeof(size_t));
		SetHighestBuild(node, level);
		SetAreaBuild(node, 1);
	}

	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::DeletingNode(SCompTreeNode<T> *node)
	{
		free(node->attributes);
		node->attributes = NULL;
		for (SCompTreeNodeListElem<T> *i = node->children.first; i != NULL; i = i->next)
			DeletingNode(i->value);
	}

	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::ChildAdded(SCompTreeNode<T> *parentNode, SCompTreeNode<T> *childNode)
	{
		I3DASSERT_ALWAYS_ET(building, "This method can be called only during the build process");
		SetAreaBuild(parentNode, GetAreatBuild(parentNode) + GetAreatBuild(childNode));
		SetHighestBuild(parentNode, std::max(GetHighestBuild(parentNode), GetHighestBuild(childNode)));
	}

	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::NodesMerged(SCompTreeNode<T> *result, SCompTreeNode<T> *mergee)
	{
		I3DASSERT_ALWAYS_ET(building, "This method can be called only during the build process");
		SetAreaBuild(result, GetAreatBuild(result) + GetAreatBuild(mergee));
		SetHighestBuild(result, std::max(GetHighestBuild(result), GetHighestBuild(mergee)));
	}

	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::BuildingFinished(SCompTreeNode<T> *node)
	{
		// allocate new space for the new attributes and copy existing ones inside
		void *newAttributes = malloc(sizeof(SCompTreeNode<T>*) + sizeof(T) + 2 * sizeof(size_t));
		void *oldAttributes = node->attributes;
		node->attributes = newAttributes;
		building = false;
		T *highest = static_cast<T*>(oldAttributes);
		SetHeight(node, *highest - node->level + 1);
		SetArea(node, *(reinterpret_cast<size_t*>(highest + 1)));
		free(oldAttributes);
		// compute and set remaining new attributes
		SetParent(node, NULL);
		size_t volume = GetArea(node);
		for (SCompTreeNodeListElem<T> *i = node->children.first; i != NULL; i = i->next)
		{
			BuildingFinished(i->value);
			SetParent(i->value, node);
			volume += GetVolume(i->value) + (GetArea(i->value) * (i->value->level - node->level - 1));
		}
		SetVolume(node, volume);
	}

	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::PrintAttrib(std::ostream &ost, const SCompTreeNode<T> *node)
	{
		// this should be called only on the node with the final attributes layout, not during the build process!
		if (typeid(T) == typeid(GRAY8))
			ost << "height: " << static_cast<size_t>(GetHeight(node));
		else
			ost << "height: " << GetHeight(node);
		std::cout << "; area: " << GetArea(node) << "; volume: " << GetVolume(node);
	}

	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::SetParent(SCompTreeNode<T> *node, SCompTreeNode<T> *parent)
	{
		*(static_cast<SCompTreeNode<T>**>(node->attributes)) = parent;
	}

	template <typename T>	
	I3D_ALGO_EXPORT SCompTreeNode<T> *CCompTreeNodeAttribHandlerPHAV<T>::GetParent(const SCompTreeNode<T> *node)
	{
		return *(static_cast<SCompTreeNode<T>**>(node->attributes));
	}
		
	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::SetHeight(SCompTreeNode<T> *node, T height)
	{
		*(reinterpret_cast<T*>(static_cast<SCompTreeNode<T>**>(node->attributes) + 1)) = height;
	}
	
	template <typename T>
	I3D_ALGO_EXPORT T CCompTreeNodeAttribHandlerPHAV<T>::GetHeight(const SCompTreeNode<T> *node)
	{
		return *(reinterpret_cast<T*>(static_cast<SCompTreeNode<T>**>(node->attributes) + 1));
	}
		
	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::SetArea(SCompTreeNode<T> *node, size_t area)
	{
		*(reinterpret_cast<size_t*>(reinterpret_cast<T*>(static_cast<SCompTreeNode<T>**>(node->attributes) + 1) + 1)) = area;
	}
		
	template <typename T>
	I3D_ALGO_EXPORT size_t CCompTreeNodeAttribHandlerPHAV<T>::GetArea(const SCompTreeNode<T> *node)
	{
		return *(reinterpret_cast<size_t*>(reinterpret_cast<T*>(static_cast<SCompTreeNode<T>**>(node->attributes) + 1) + 1));
	}
		
	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::SetVolume(SCompTreeNode<T> *node, size_t volume)
	{
		*(reinterpret_cast<size_t*>(reinterpret_cast<T*>(static_cast<SCompTreeNode<T>**>(node->attributes) + 1) + 1) + 1) = volume;
	}

	template <typename T>
	I3D_ALGO_EXPORT size_t CCompTreeNodeAttribHandlerPHAV<T>::GetVolume(const SCompTreeNode<T> *node)
	{
		return *(reinterpret_cast<size_t*>(reinterpret_cast<T*>(static_cast<SCompTreeNode<T>**>(node->attributes) + 1) + 1) + 1);
	}

	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::SetHighestBuild(SCompTreeNode<T> *node, T highest)
	{
		*(static_cast<T*>(node->attributes)) = highest;
	}

	template <typename T>
	I3D_ALGO_EXPORT T CCompTreeNodeAttribHandlerPHAV<T>::GetHighestBuild(const SCompTreeNode<T> *node)
	{
		return *(static_cast<T*>(node->attributes));
	}

	template <typename T>
	I3D_ALGO_EXPORT void CCompTreeNodeAttribHandlerPHAV<T>::SetAreaBuild(SCompTreeNode<T> *node, size_t area)
	{
		*(reinterpret_cast<size_t*>(static_cast<T*>(node->attributes) + 1)) = area;
	}

	template <typename T>
	I3D_ALGO_EXPORT size_t CCompTreeNodeAttribHandlerPHAV<T>::GetAreatBuild(const SCompTreeNode<T> *node)
	{
		return *(reinterpret_cast<size_t*>(static_cast<T*>(node->attributes) + 1));
	}

	template I3D_ALGO_EXPORT class CCompTreeNodeAttribHandlerPHAV<i3d::GRAY8>;
	template I3D_ALGO_EXPORT class CCompTreeNodeAttribHandlerPHAV<i3d::GRAY16>;

	//--------------------------------------------------------------------------
	// Component three methods
	//--------------------------------------------------------------------------

	template <typename T>
	I3D_ALGO_EXPORT SCompTree<T>::SCompTree(CCompTreeNodeAttribHandler<T> *atribHandler)
	{
		root = NULL;
		mapping = NULL;
		this->atribHandler = atribHandler;
	}

	template <typename T>
	I3D_ALGO_EXPORT SCompTree<T>::~SCompTree()
	{
		if (atribHandler && root)
			atribHandler->DeletingNode(root);
		if (root)
			delete root;
		if (mapping)
			delete[] mapping;
		if (atribHandler)
			delete atribHandler;

	}

	template I3D_ALGO_EXPORT struct SCompTree<i3d::GRAY8>;
	template I3D_ALGO_EXPORT struct SCompTree<i3d::GRAY16>;

	//--------------------------------------------------------------------------
	// Component three stream output
	//--------------------------------------------------------------------------

	template <typename T> 
	I3D_ALGO_EXPORT void PrintSCompTreeNode(std::ostream &ost, const SCompTreeNode<T> *compTreeNode, CCompTreeNodeAttribHandler<T> *atribHandler, size_t level)
	{
		for (size_t i = 0; i < level; ++i)
			ost << "|";
		ost << "-";
		ost << "(id: " << compTreeNode << "; level: ";
		if (typeid(T) == typeid(GRAY8))
			ost  << static_cast<size_t>(compTreeNode->level);
		else
			ost << compTreeNode->level;
		if (atribHandler)
		{
			std::cout << "; ";
			atribHandler->PrintAttrib(ost, compTreeNode);
		}
		ost << ")" << std::endl;
		for (SCompTreeNodeListElem<T> *i = compTreeNode->children.first; i != NULL; i = i->next)
			PrintSCompTreeNode(ost, i->value, atribHandler, level + 1);
	}

	template <typename T>
	I3D_ALGO_EXPORT std::ostream& operator<<(std::ostream &ost, const SCompTree<T> &compTree)
	{
		return ost << &compTree;
	}

	template <typename T>
	I3D_ALGO_EXPORT std::ostream& operator<<(std::ostream &ost, const SCompTree<T> *compTree)
	{
		PrintSCompTreeNode(ost, compTree->root, compTree->atribHandler, 0);
		return ost;
	}

	template I3D_ALGO_EXPORT std::ostream& operator<<(std::ostream &ost, const SCompTree<i3d::GRAY8> &compTreeNode);
	template I3D_ALGO_EXPORT std::ostream& operator<<(std::ostream &ost, const SCompTree<i3d::GRAY16> &compTreeNode);
	template I3D_ALGO_EXPORT std::ostream& operator<<(std::ostream &ost, const SCompTree<i3d::GRAY8> *compTreeNode);
	template I3D_ALGO_EXPORT std::ostream& operator<<(std::ostream &ost, const SCompTree<i3d::GRAY16> *compTreeNode);


	//--------------------------------------------------------------------------
	// Image voxels sorting related functions and structures
	//--------------------------------------------------------------------------

	/// structure used in the CompTreeSort function while sorting the indices according to the decreasing value of the voxel intensities (increasingly for the same voxel intensities) using the std library
	template<typename T>
	struct SCompTreeSortStdLibSorter
	{
		const T *intensities;
		bool operator()(const size_t &value1, const size_t &value2)
		{
			return ((intensities[value1] > intensities[value2]) || (intensities[value1] == intensities[value2] && value1 < value2));
		}
	};

	/// returns array of indices into the given image that allows scanning of the voxels in decreasing order
	template<typename T>
	static size_t* CompTreeSort(const Image3d<T> &input)
	{
		size_t inputSize = input.GetImageSize();
		size_t *result = new size_t[inputSize];

		// choose the best sorting algorithm for the selected data type
		const std::type_info &typeOfT = typeid(T);
		if (typeOfT == typeid(GRAY8) || typeOfT == typeid(GRAY16))
		{
			// for small and countable ranges we use counting sort implementation
			// create histogram
			T minTValue = std::numeric_limits<T>::min();
			size_t rangeSize = static_cast<size_t>(std::numeric_limits<T>::max() - minTValue) + 1;
			size_t *histogram = new size_t[rangeSize];
			memset(histogram, 0, rangeSize * sizeof(size_t));
			for (size_t i = 0; i < inputSize; ++i)
				histogram[input.GetVoxel(i) - minTValue]++;
			// sum up the histogram - get the mapping of the intensities in the decreasing order into the index ranges
			for (size_t i = rangeSize - 1; i > 0; --i)
				histogram[i - 1] += histogram[i];
			// go through the whole image and write indices according to the values in histogram (going bacwards because of the structure of the indices represented in the histogram and because we need the indices sorted decreasing by level but increasing by value of the index)
			for (size_t i = inputSize; i > 0; --i)
				result[--histogram[input.GetVoxel(i - 1) - minTValue]] = i - 1;
			// free what we do not need any more
			delete[] histogram;
		}
		else
		{
			// for other data types we use simple std library sort, which is good enough these days
			for (size_t i = 0; i < inputSize; ++i)
				result[i] = i;
			SCompTreeSortStdLibSorter<T> compTreeSortStdLibSorter;
			compTreeSortStdLibSorter.intensities = input.GetFirstVoxelAddr();
			std::sort(result, result + inputSize, compTreeSortStdLibSorter);
		}
		return result;
	}

	//--------------------------------------------------------------------------
	// Component tree building algorithm
	//--------------------------------------------------------------------------

	template <typename T>
	I3D_ALGO_EXPORT void BuildComponentTree(const Image3d<T> &input, SCompTree<T> &componentTree, const Neighbourhood &neighbourhood)
	{
		NeighbourhoodWalker nbhWalker(input.GetSize(), neighbourhood, false);
		BuildComponentTree(input, componentTree, nbhWalker);
	}

	template I3D_ALGO_EXPORT void BuildComponentTree(const Image3d<i3d::GRAY8> &input, SCompTree<i3d::GRAY8> &componentTree, const Neighbourhood &neighbourhood);
	template I3D_ALGO_EXPORT void BuildComponentTree(const Image3d<i3d::GRAY16> &input, SCompTree<i3d::GRAY16> &componentTree, const Neighbourhood &neighbourhood);

	template <typename T>
	I3D_ALGO_EXPORT void BuildComponentTree(const Image3d<T> &input, SCompTree<T> &componentTree, NeighbourhoodWalker &nbhWalker)
	{
		I3DASSERT_ALWAYS_ET(componentTree.root == NULL && componentTree.mapping == NULL, "Given component tree is already initialized.");
		// initialize useful data structures
		size_t imageSize = input.GetImageSize();
		size_t *scanningOrder = CompTreeSort(input); // order in which we process the voxels - from the highest ones (voxel on the same level ale sorted according to their index - this allows us to distinguish already processed or unprocessed neighbours)
		CCompTreeUnionFind *parTree = new CCompTreeUnionFind(imageSize); // used to represent partialy constructed trees (returns representant of the tree)
		CCompTreeUnionFind *parNode = new CCompTreeUnionFind(imageSize); // used to represent partialy constructed nodes (returns representant of the node)
		size_t *lowestNode = new size_t[imageSize]; // for each partial tree returns one node for the current level
		SCompTreeNode<T> **nodes = new SCompTreeNode<T>*[imageSize]; // nodes for the future component tree
		for (size_t i = 0; i < imageSize; ++i)
		{	
			lowestNode[i] = i;
			T level = input.GetVoxel(i);
			nodes[i] = new SCompTreeNode<T>(level);
			if (componentTree.atribHandler)
				componentTree.atribHandler->NodeCreated(nodes[i], level);
		}
		
		// process voxel by voxel from the highest function values
		for (size_t i = 0; i < imageSize; ++i)
		{
			// for currently processed voxel, get the partial tree (representant) and the node (representant) for the current level
			size_t curVoxel = scanningOrder[i];
			size_t curTree = parTree->Find(curVoxel);
			size_t curNode = parNode->Find(lowestNode[curTree]);
			// check all the neighbours of the current voxel that has higher or the same intensity and has already been processed
			for (nbhWalker.Begin(curVoxel); nbhWalker.Valid(); nbhWalker.Next())
			{
				size_t adjVoxel = nbhWalker.GetIndex();
				if ((input.GetVoxel(adjVoxel) < input.GetVoxel(curVoxel)) || (input.GetVoxel(adjVoxel) == input.GetVoxel(curVoxel) && adjVoxel > curVoxel))
					continue;
				// for currently processed neighbour, get the partial tree (representant) and the node (representant) for the neighbour's level
				size_t adjTree = parTree->Find(adjVoxel);
				size_t adjNode = parNode->Find(lowestNode[adjTree]);
				// we are interested in the cases when the currently processed voxel and it's relevant neighbour are from the different nodes (otherwise there is no new information for the component tree:)
				if (curNode != adjNode)
				{
					// now it is necessary to distinguish the cases when processed nodes are on the same level or adjacent node is on the strictly higher level
					if (nodes[curNode]->level == nodes[adjNode]->level)
					{
						// adjacent node is on the same level, so we need to merge these two nodes because they are the both parts of the same component - we get the result node and a node for deletion that is not needed any more
						size_t mrgNode = parNode->Link(curNode, adjNode);
						size_t delNode;
						if (mrgNode == adjNode)
							delNode = curNode;
						else
							delNode = adjNode;
						nodes[mrgNode]->children.Merge(nodes[delNode]->children);
						if (componentTree.atribHandler)
							componentTree.atribHandler->NodesMerged(nodes[mrgNode], nodes[delNode]);
						curNode = mrgNode;
						// non-canonical nodes (nodes which are not part of the partial tree and are parts of some somponent of their level) are not needed any more
						if (componentTree.atribHandler)
							componentTree.atribHandler->DeletingNode(nodes[delNode]);
						delete nodes[delNode];
						nodes[delNode] = NULL;
					}
					else
					{
						// adjacent node is strictly higher so current lower node becomes a parent of the higher node
						nodes[curNode]->children.PushBack(nodes[adjNode]);
						if (componentTree.atribHandler)
							componentTree.atribHandler->ChildAdded(nodes[curNode], nodes[adjNode]);
					}
					curTree = parTree->Link(adjTree, curTree);
					lowestNode[curTree] = curNode;
				}
			}
		}
		componentTree.root = nodes[lowestNode[parTree->Find(parNode->Find(0))]];
		// free some resources that are not needed any more
		delete[] lowestNode;
		delete parTree;
		delete[] scanningOrder;
		// construct the component mapping
		componentTree.mapping = new SCompTreeNode<T>*[imageSize];
		for (size_t i = 0; i < imageSize; ++i)
		{
			componentTree.mapping[i] = nodes[parNode->Find(i)];
			I3DASSERT_BT(componentTree.mapping[i], "Component mapping contains an empty component, something is wrong with node deletion.");
		}
		// free remainint allocated resources that are not needed any more
		delete[] nodes;
		delete parNode;
		// use attribute manager to form the final attributes
		if (componentTree.atribHandler)
			componentTree.atribHandler->BuildingFinished(componentTree.root);
	}

	template I3D_ALGO_EXPORT void BuildComponentTree(const Image3d<i3d::GRAY8> &input, SCompTree<i3d::GRAY8> &componentTree, NeighbourhoodWalker &nbhWalker);
	template I3D_ALGO_EXPORT void BuildComponentTree(const Image3d<i3d::GRAY16> &input, SCompTree<i3d::GRAY16> &componentTree, NeighbourhoodWalker &nbhWalker);

} // namespace i3d
