/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Freere
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/** 
 * \file mpeg7ehd.cc
   \brief 2D & 3D MPEG-7 EHD features.
	
	This file contains functions for calculating 2D/3D MPEG-7 EHD features.

	[1] B. S. Manjunath, P. Salembier, and T. Sikora, editors. Introduction to
	MPEG-7: Multimedia Content Description Interface. Wiley & Sons, Inc.,
	New York, USA, Apr. 2002.
	
	[2] T. Majtner, D. Svoboda. Texture Analysis Using 3D Gabor Features and
	3D MPEG-7 Edge Histogram Descriptor in Fluorescence Microscopy. In
	4th International Conference on 3D Imaging (IC3D). Los Alamitos, California,
	USA: IEEE Computer Society, 2014. Pages 17, ISBN 978-1-4799-8023-9.
	
	\author Tomas Majtner <172786@mail.muni.cz>
	\date 2015
*/

#include "transform.h"
#include "descriptors/mpeg7ehd.h"
#include <fstream>
#include <sstream>
#include <iostream>
#include <math.h>
#include <stdio.h>

using namespace std;

namespace i3d {


template <class T> void Compute2DEHD(Image3d<T> &img, vector<double> &final_vector)
{
	// block size from [1]
	int block_size = (int)floor(sqrt(img.GetSizeX()*img.GetSizeY()/1100.0));
	block_size = std::max(2, block_size);
	if (block_size%2 != 0) block_size--;

	int hist_bins[4][4][5];    // 5 bins for each 4x4 subimage
	int global_hist[5];

	// clear the histogram
	for (int i = 0; i < 5; i++) {
		for (int j = 0; j < 4; j++) {
			for (int k = 0; k < 4; k++) {
				hist_bins[k][j][i] = 0;				
			}
		}
	}
	for (int i = 0; i < 5; i++) global_hist[i] = 0;

	// 4 parts of subimage: top-left, top-right, bottom-left, bottom-right parts
	double topL, topR, bottomL, bottomR;
	int ii, jj;			 // indexing variables for later histogram forming
	int threshold = 11;  // threshold value from [1]

	for (size_t y = 0; y < img.GetSizeY()-block_size; y = y+block_size) {
		for (size_t x = 0; x < img.GetSizeX()-block_size; x = x+block_size) {

			// sum the top left part of subimage
			topL = 0; topR = 0; bottomL = 0; bottomR = 0;
			for (size_t i = x; i < x+(block_size/2); i++) {
				for (size_t j = y; j < y+(block_size/2); j++) {
					topL += img.GetVoxel(i,j,0);
				}
			}
			topL = (double)topL / block_size;

			// sum the top right part of subimage
			for (size_t i = x+(block_size/2); i < std::min(x+block_size, img.GetSizeX()); i++) {
				for (size_t j = y; j < y+(block_size/2); j++) {
					topR += img.GetVoxel(i,j,0);
				}
			}
			topR = (double)topR / block_size;

			// sum the bottom left part of subimage
			for (size_t i = x; i < x+(block_size/2); i++) {
				for (size_t j = y+(block_size/2); j < std::min(y+block_size, img.GetSizeY()); j++) {
					bottomL += img.GetVoxel(i,j,0);
				}
			}
			bottomL = (double)bottomL / block_size;

			// sum the bottom right part of subimage
			for (size_t i = x+(block_size/2); i < std::min(x+block_size, img.GetSizeX()); i++) {
				for (size_t j = y+(block_size/2); j < std::min(y+block_size, img.GetSizeY()); j++) {
					bottomR += img.GetVoxel(i,j,0);
				}
			}
			bottomR = (double)bottomR / block_size;			

			// finding the max value using derived sums
			double max_val = std::max(std::abs(topL + bottomL - topR - bottomR), 
				             std::max(std::abs(topL + topR - bottomL - bottomR), 
							 std::max(std::abs(sqrt(2.0)*topL - sqrt(2.0)*bottomR), 
							 std::max(std::abs(sqrt(2.0)*topR - sqrt(2.0)*bottomL), 
							     std::abs(2*topL + 2*bottomR - 2*topR - 2*bottomL)))));

			// comparing max value with the threshold from [1]
			if (max_val > threshold)
			{
				ii = (int)floor(x / (img.GetSizeX() / 4.0));
				jj = (int)floor(y / (img.GetSizeY() / 4.0));			
				if (std::abs(topL + bottomL - topR - bottomR) >= max_val) { hist_bins[ii][jj][0]++; global_hist[0]++; }
				if (std::abs(topL + topR - bottomL - bottomR) >= max_val) { hist_bins[ii][jj][1]++; global_hist[1]++; }
				if (std::abs(sqrt(2.0)*topL - sqrt(2.0)*bottomR) >= max_val) { hist_bins[ii][jj][2]++; global_hist[2]++; }
				if (std::abs(sqrt(2.0)*topR - sqrt(2.0)*bottomL) >= max_val) { hist_bins[ii][jj][3]++; global_hist[3]++; }
				if (std::abs(2*topL + 2*bottomR - 2*topR - 2*bottomL) >= max_val) { hist_bins[ii][jj][4]++; global_hist[4]++; }
			}
		}
	}

	// printing local edges
	for (int j = 0; j < 4; j++) {
		for (int k = 0; k < 4; k++) {
			for (int i = 0; i < 5; i++) {
				final_vector.push_back(hist_bins[k][j][i]);
			}
		}
	}

	// printing semilocal edges
	int temp = 0;
	for (int j = 0; j < 4; j++) {
		for (int i = 0; i < 5; i++) {
			for (int k = 0; k < 4; k++) {
				temp += hist_bins[k][j][i];
			}
			final_vector.push_back(temp);
			temp = 0;
		}
	}
	for (int j = 0; j < 4; j++) {
		for (int i = 0; i < 5; i++) {
			for (int k = 0; k < 4; k++) {
				temp += hist_bins[j][k][i];
			}
			final_vector.push_back(temp);
			temp = 0;
		}
	}
	for (int i = 0; i < 5; i++) {
		for (int j = 1; j < 3; j++) {
			for (int k = 1; k < 3; k++) {
				temp += hist_bins[k][j][i];
			}			
		}
		final_vector.push_back(temp);
		temp = 0;
	}

	// printing global edges
	for (int i = 0; i < 4; i++) final_vector.push_back(global_hist[i]);
	final_vector.push_back(global_hist[4]);
}


template <class T> void Compute3DEHD(Image3d<T> &img, vector<double> &final_vector)
{
	// *** Resample To Desired Resolution ***
	Vector3d<float> res = img.GetResolution().GetRes();
	if ((res.x != res.y) || (res.x != res.z))
	{
		 float max_res = std::max(std::max(res.x, res.y), res.z);
		 Vector3d<float> new_res = max_res;
		 ResampleToDesiredResolution(img, new_res, LANCZOS);
	}

	// block size from [2] 
	int block_size = (int)floor(pow(pow(std::min(img.GetSizeX(), 
		std::min(img.GetSizeY(), img.GetSizeZ())), 3.0)/140000.0, 1.0/3.0));
	block_size = std::max(2, block_size);
	if (block_size%2 != 0) block_size--;

	int hist_bins[4][4][4][10];    // 10 bins for each 4x4x4 subimage
	int global_hist[10];

	// clear the histogram
	for (int i = 0; i < 10; i++) {
		for (int j = 0; j < 4; j++) {
			for (int k = 0; k < 4; k++) {
				for (int l = 0; l < 4; l++) {
					hist_bins[l][k][j][i] = 0;
				}
			}
		}
	}
	for (int i = 0; i < 10; i++) global_hist[i] = 0;
	
	// 4 parts of subimage for front top-left, top-right, bottom-left, bottom-right parts
	// 4 parts of subimage for back top-left, top-right, bottom-left, bottom-right parts
	double topleftF, topleftB, toprightF, toprightB;
	double bottomleftF, bottomleftB, bottomrightF, bottomrightB;
	int ii, jj, kk;      // indexing variables for later histogram forming
	int threshold = 20;  // threshold value from [2]

	for (size_t z = 0; z < img.GetSizeZ()-block_size; z = z+block_size) {
		for (size_t y = 0; y < img.GetSizeY()-block_size; y = y+block_size) {
			for (size_t x = 0; x < img.GetSizeX()-block_size; x = x+block_size) {
		
				topleftF = 0; topleftB = 0; toprightF = 0; toprightB = 0;
				bottomleftF = 0; bottomleftB = 0; bottomrightF = 0; bottomrightB = 0;

				// sum the front top left part of subimage
				for (size_t i = x; i < x+(block_size/2); i++) {
					for (size_t j = y; j < y+(block_size/2); j++) {
						for (size_t k = z; k < z+(block_size/2); k++) {
							topleftF += img.GetVoxel(i,j,k);
						}
					}
				}
				topleftF = (double)topleftF / block_size;

				// sum the front top right part of subimage
				for (size_t i = x+(block_size/2); i < std::min(x+block_size, img.GetSizeX()); i++) {
					for (size_t j = y; j < y+(block_size/2); j++) {
						for (size_t k = z; k < z+(block_size/2); k++) {
							toprightF += img.GetVoxel(i,j,k);
						}
					}
				}
				toprightF = (double)toprightF / block_size;

				// sum the front bottom left part of subimage
				for (size_t i = x; i < x+(block_size/2); i++) {
					for (size_t j = y+(block_size/2); j < std::min(y+block_size, img.GetSizeY()); j++) {
						for (size_t k = z; k < z+(block_size/2); k++) {
							bottomleftF += img.GetVoxel(i,j,k);
						}
					}
				}
				bottomleftF = (double)bottomleftF / block_size;

				// sum the front bottom right part of subimage
				for (size_t i = x+(block_size/2); i < std::min(x+block_size, img.GetSizeX()); i++) {
					for (size_t j = y+(block_size/2); j < std::min(y+block_size, img.GetSizeY()); j++) {
						for (size_t k = z; k < z+(block_size/2); k++) {
							bottomrightF += img.GetVoxel(i,j,k);
						}
					}
				}
				bottomrightF = (double)bottomrightF / block_size;

				// sum the back top left part of subimage
				for (size_t i = x; i < x+(block_size/2); i++) {
					for (size_t j = y; j < y+(block_size/2); j++) {
						for (size_t k = z+(block_size/2); k < std::min(z+block_size, img.GetSizeZ()); k++) {
							topleftB += img.GetVoxel(i,j,k);
						}
					}
				}
				topleftB = (double)topleftB / block_size;

				// sum the back top right part of subimage
				for (size_t i = x+(block_size/2); i < std::min(x+block_size, img.GetSizeX()); i++) {
					for (size_t j = y; j < y+(block_size/2); j++) {
						for (size_t k = z+(block_size/2); k < std::min(z+block_size, img.GetSizeZ()); k++) {
							toprightB += img.GetVoxel(i,j,k);
						}
					}
				}
				toprightB = (double)toprightB / block_size;

				// sum the back bottom left part of subimage
				for (size_t i = x; i < x+(block_size/2); i++) {
					for (size_t j = y+(block_size/2); j < std::min(y+block_size, img.GetSizeY()); j++) {
						for (size_t k = z+(block_size/2); k < std::min(z+block_size, img.GetSizeZ()); k++) {
							bottomleftB += img.GetVoxel(i,j,k);
						}
					}
				}
				bottomleftB = (double)bottomleftB / block_size;

				// sum the back bottom right part of subimage
				for (size_t i = x+(block_size/2); i < std::min(x+block_size, img.GetSizeX()); i++) {
					for (size_t j = y+(block_size/2); j < std::min(y+block_size, img.GetSizeY()); j++) {
						for (size_t k = z+(block_size/2); k < std::min(z+block_size, img.GetSizeZ()); k++) {
							bottomrightB += img.GetVoxel(i,j,k);
						}
					}
				}
				bottomrightB = (double)bottomrightB / block_size;

				// temp values for finding the max values using derived sums
				double val1 = std::abs(topleftF + bottomleftF + topleftB + bottomleftB - 
					toprightF - bottomrightF - toprightB - bottomrightB);
				double val2 = std::abs(bottomleftF + bottomrightF + bottomleftB + bottomrightB -
					topleftF - toprightF - topleftB - toprightB);
				double val3 = std::abs(topleftB + toprightB + bottomleftB + bottomrightB -
					topleftF - toprightF - bottomleftF - bottomrightF);
				double val4 = std::abs(sqrt(2.0)*bottomleftF + sqrt(2.0)*bottomleftB -
					sqrt(2.0)*toprightF - sqrt(2.0)*toprightB);
				double val5 = std::abs(sqrt(2.0)*bottomrightF + sqrt(2.0)*bottomrightB -
					sqrt(2.0)*topleftF - sqrt(2.0)*topleftB);
				double val6 = std::abs(sqrt(2.0)*bottomleftB + sqrt(2.0)*bottomrightB -
					sqrt(2.0)*topleftF - sqrt(2.0)*toprightF);
				double val7 = std::abs(sqrt(2.0)*bottomleftF + sqrt(2.0)*bottomrightF -
					sqrt(2.0)*topleftB - sqrt(2.0)*toprightB);
				double val8 = std::abs(sqrt(2.0)*topleftB + sqrt(2.0)*bottomleftB -
					sqrt(2.0)*toprightF - sqrt(2.0)*bottomrightF);
				double val9 = std::abs(sqrt(2.0)*toprightB + sqrt(2.0)*bottomrightB -
					sqrt(2.0)*topleftF - sqrt(2.0)*bottomleftF);
				double val0 = std::abs(topleftF + bottomrightF + toprightB + bottomleftB - 
					topleftB - bottomrightB - toprightF - bottomleftF);

				// determining the max value
				double max_val = std::max(val1, std::max(val2, std::max(val3, std::max(val4, std::max(val5, 
					std::max(val6, std::max(val7, std::max(val8, std::max(val9, val0)))))))));

				// comparing max value with the threshold from [2]
				if (max_val > threshold)
				{
					// indexing histogram
					ii = (int)floor(x / (img.GetSizeX() / 4.0));
					jj = (int)floor(y / (img.GetSizeY() / 4.0));
					kk = (int)floor(z / (img.GetSizeZ() / 4.0));

					if (val1 >= max_val) { hist_bins[ii][jj][kk][1]++; global_hist[1]++; }
					if (val2 >= max_val) { hist_bins[ii][jj][kk][2]++; global_hist[2]++; }
					if (val3 >= max_val) { hist_bins[ii][jj][kk][3]++; global_hist[3]++; }
					if (val4 >= max_val) { hist_bins[ii][jj][kk][4]++; global_hist[4]++; }
					if (val5 >= max_val) { hist_bins[ii][jj][kk][5]++; global_hist[5]++; }
					if (val6 >= max_val) { hist_bins[ii][jj][kk][6]++; global_hist[6]++; }
					if (val7 >= max_val) { hist_bins[ii][jj][kk][7]++; global_hist[7]++; }
					if (val8 >= max_val) { hist_bins[ii][jj][kk][8]++; global_hist[8]++; }
					if (val9 >= max_val) { hist_bins[ii][jj][kk][9]++; global_hist[9]++; }
					if (val0 >= max_val) { hist_bins[ii][jj][kk][0]++; global_hist[0]++; }
				}
			}
		}
	}

	// printing local edges
	for (int l = 0; l < 4; l++) {
		for (int k = 0; k < 4; k++) {
			for (int j = 0; j < 4; j++) {		
				for (int i = 0; i < 10; i++) {
					final_vector.push_back(hist_bins[l][k][j][i]);
				}
			}
		}
	}

	// printing semilocal edges
	int temp = 0;
	for (int j = 0; j < 4; j++) {
		for (int i = 0; i < 10; i++) {
			for (int l = 0; l < 4; l++) {
				for (int k = 0; k < 4; k++) {
					temp += hist_bins[l][k][j][i];
				}
			}
			final_vector.push_back(temp);
			temp = 0;
		}
	}
	for (int j = 0; j < 4; j++) {
		for (int i = 0; i < 10; i++) {
			for (int l = 0; l < 4; l++) {
				for (int k = 0; k < 4; k++) {
					temp += hist_bins[l][j][k][i];
				}
			}
			final_vector.push_back(temp);
			temp = 0;
		}
	}
	for (int j = 0; j < 4; j++) {
		for (int i = 0; i < 10; i++) {
			for (int l = 0; l < 4; l++) {
				for (int k = 0; k < 4; k++) {
					temp += hist_bins[j][l][k][i];
				}
			}
			final_vector.push_back(temp);
			temp = 0;
		}
	}
	for (int i = 0; i < 10; i++) {
		for (int j = 1; j < 3; j++) {
			for (int k = 1; k < 3; k++) {
				for (int l = 1; l < 3; l++) {
					temp += hist_bins[l][k][j][i];
				}
			}			
		}
		final_vector.push_back(temp);
		temp = 0;
	}

	for (int i = 0; i < 9; i++) final_vector.push_back(global_hist[i]);
	final_vector.push_back(global_hist[9]);
}

//-----------------------------------------------------------------------
// explicit instantiation
//-----------------------------------------------------------------------

template I3D_ALGO_EXPORT void Compute2DEHD<GRAY8>(Image3d<GRAY8> &, vector<double> &);
template I3D_ALGO_EXPORT void Compute2DEHD<GRAY16>(Image3d<GRAY16> &, vector<double> &);

template I3D_ALGO_EXPORT void Compute3DEHD<GRAY8>(Image3d<GRAY8> &, vector<double> &);
template I3D_ALGO_EXPORT void Compute3DEHD<GRAY16>(Image3d<GRAY16> &, vector<double> &);

}
