/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2015   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Freere
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/** 
 * \file rsurf.cc
   \brief 2D & 3D RSurf features.
	
	This file contains functions for calculating 2D/3D RSurf features.

	[1] T. Majtner, R. Stoklasa, D. Svoboda. RSurf  The Efficient Texture-based
	Descriptor for Fluorescence Microscopy Images of HEp-2 Cells. In 22nd
	International Conference on Pattern Recognition (ICPR). Los Alamitos, California,
	USA: IEEE Computer Society, 2014. Pages 11941199, ISBN 978-1-
	4799-5208-3.
	
	\author Tomas Majtner <172786@mail.muni.cz>
	\author Roman Stoklasa <139873@mail.muni.cz>
	\date 2015
*/

#include <fstream>
#include <string> 
#include <iostream>
#include <algorithm>
#include <math.h>

#include "transform.h"
#include "filters.h"
#include "descriptors/rsurfslopesdivider.h"
#include "descriptors/rsurf.h"

// default values from [1]
#define LEN 16
#define ANGLES "0,45,90,135"
#define ANGLES3D "0,45,90,135"
#define SIGMA 0.5
#define SKIP_FACTOR 1
#define MAX_LENGTH 40
#define MAX_HEIGHT 250

using namespace std;

namespace i3d {

int totalSlopesLength = 0;
vector <vector<float> > UsedOrientations;

bool AreSame(double a, double b)
{
    return fabs(a - b) < 0.01;
}

template <class T> vector<vector<float> > GetLinesFrom3D(Image3d<T> &img, int alpha_deg, int beta_deg)
{
	// getting size of an image
	int Nx = img.GetSizeX();
	int Ny = img.GetSizeY();
	int Nz = img.GetSizeZ();

	vector<float> line;
	vector<vector<float> > result;

	// to radians
	double alpha = alpha_deg * M_PI / 180;
	double beta = beta_deg * M_PI / 180;

	// imaginary end point
	int x1 = (beta_deg == 90 || alpha_deg == 180) ? 0 : (int)round(Nx*Ny*sin(alpha)*cos(beta));
	int y1 = (beta_deg == 180 || alpha_deg == 180) ? 0 : (int)round(Nx*Ny*sin(alpha)*sin(beta));
	int z1 = (alpha_deg == 90) ? 0 : (int)round(Nx*Ny*cos(alpha));
	int ax = 2*std::abs(x1);
	int ay = 2*std::abs(y1);
	int az = 2*std::abs(z1);

	#ifdef I3D_DEBUG
		for (double y = 0; y < 181; y=y+15) {
			for (double x = 0; x < 181; x=x+15) {
					double alpha = x * M_PI / 180;
					double beta = y * M_PI / 180;

				int x1 = (y == 90 || x == 180) ? 0 : (int)round(Nx*Ny*sin(alpha)*cos(beta));
				int y1 = (y == 180 || x == 180) ? 0 : (int)round(Nx*Ny*sin(alpha)*sin(beta));
				int z1 = (x == 90) ? 0 : (int)round(Nx*Ny*cos(alpha));
				cout << x1 << "\t" << y1 << "\t" << z1 << endl;
			}
		}

		int suma = 0;
		Image3d<T> test;
		test.MakeRoom(img.GetSizeX(), img.GetSizeY(), img.GetSizeZ());
		test.SetAllVoxels(0);
	#endif

	// x dominant
	if (ax >= std::max(ay, az))							
	{
		for (int z = -Nz; z < 2*Nz; z=z+SKIP_FACTOR)
		{
			for (int y = -Ny; y < 2*Ny; y=y+SKIP_FACTOR)
			{
				int x0 = (beta_deg < 90) ? 0 : Nx-1;
				int y0 = y;
				int z0 = z;

				signed short sx = (x1 > x0) ? 1 : (x1 < x0) ? -1 : 0;
				signed short sy = (y1 > y0) ? 1 : (y1 < y0) ? -1 : 0;
				signed short sz = (z1 > z0) ? 1 : (z1 < z0) ? -1 : 0;

				int yd = ay - ax/2;
				int zd = az - ax/2;
				while (1)
				{
					if (x0 >= 0 && x0 < Nx && y0 >= 0 && y0 < Ny && z0 >= 0 && z0 < Nz)
					{
						line.push_back(img.GetVoxel(x0, y0, z0));
					}

					if ((beta_deg < 90 && x0 > Nx) || (beta_deg > 90 && x0 < 0) || (beta_deg == 90)) break;
					if (yd >= 0) { y0 += sy; yd -= ax; }  // move along y
					if (zd >= 0) { z0 += sz; zd -= ax; }  // move along z
					x0 += sx;							  // move along x
					yd += ay;
					zd += az;
				}
				result.push_back(line);
				line.clear();
			}
		}
	}

	// y dominant
	else if (ay >= std::max(ax, az))						
	{
		for (int z = -Nz; z < 2*Nz; z=z+SKIP_FACTOR)
		{
			for (int x = -Nx; x < 2*Nx; x=x+SKIP_FACTOR)
			{
				int x0 = x;
				int y0 = 0;
				int z0 = z;

				signed short sx = (x1 > x0) ? 1 : (x1 < x0) ? -1 : 0;
				signed short sy = (y1 > y0) ? 1 : (y1 < y0) ? -1 : 0;
				signed short sz = (z1 > z0) ? 1 : (z1 < z0) ? -1 : 0;

				int xd = ax - ay/2;
				int zd = az - ay/2;      
				while (1)
				{
					if (x0 >= 0 && x0 < Nx && y0 >= 0 && y0 < Ny && z0 >= 0 && z0 < Nz) {
						line.push_back(img.GetVoxel(x0, y0, z0));
					}
					
					if (y0 > Ny) break;
					if (xd >= 0) { x0 += sx; xd -= ay; }  // move along x
					if (zd >= 0) { z0 += sz; zd -= ay; }  // move along z
					y0 += sy;					  		  // move along y
					xd += ax;
					zd += az;
				}
				result.push_back(line);
				line.clear();
			}
		}
	}

	// z dominant
	else if (az >= std::max(ax, ay))						
	{
		for (int y = -Ny; y < 2*Ny; y=y+SKIP_FACTOR)
		{
			for (int x = -Nx; x < 2*Nx; x=x+SKIP_FACTOR)
			{
				int x0 = x;
				int y0 = y;
				int z0 = (alpha_deg < 90) ? 0 : Nz-1;

				signed short sx = (x1 > x0) ? 1 : (x1 < x0) ? -1 : 0;
				signed short sy = (y1 > y0) ? 1 : (y1 < y0) ? -1 : 0;
				signed short sz = (z1 > z0) ? 1 : (z1 < z0) ? -1 : 0;

				int xd = ax - az/2;
				int yd = ay - az/2;
				while (1)
				{
					if (x0 >= 0 && x0 < Nx && y0 >= 0 && y0 < Ny && z0 >= 0 && z0 < Nz) 
					{
						line.push_back(img.GetVoxel(x0, y0, z0));
					}

					if ((alpha_deg < 90 && z0 > Nz) || (alpha_deg > 90 && z0 < 0) || (alpha_deg == 90)) break;
					if (xd >= 0) { x0 += sx; xd -= az; }  // move along x
					if (yd >= 0) { y0 += sy; yd -= az; }  // move along y
					z0 += sz;							  // move along z
					xd += ax;
					yd += ay;
				}
				result.push_back(line);
				line.clear();
			}
		}
	}			
	return result;
}

template <class T> vector<vector<float> > GetLines(Image3d<T> &img, int degrees)
{
	// getting size of an image
	int Nx = img.GetSizeX();
	int Ny = img.GetSizeY();

	// to radians
	double angle = (degrees <= 90) ? (degrees * M_PI / 180) : ((180-degrees) * M_PI / 180);

	// imaginary end point
	int dx = (int)floor(Nx*Ny*sin(angle)); 
	int dy = (int)floor(Nx*Ny*cos(angle));	

	vector<float> line;
	vector<vector<float> > result;

	if (degrees >= 45 && degrees < 135)
	{
		for (int y = -std::max(Nx,Ny); y < std::max(Nx,Ny); y++)
		{
			int x0 = (degrees <= 90) ? 0 : Nx-1;
			int y0 = y;
			int err = dx-dy;

			while ((x0 < 2*Nx) && (y0 < 2*Ny) && (x0 >= -Nx))
			{
				if ((x0 < Nx) && (y0 < Ny) && (x0 >= 0) && (y0 >= 0))
				{
					line.push_back(img.GetVoxel(x0, y0, 0));
				}

				int e2 = 2*err;
				if (e2 > -dy) 
				{
					err = err - dy;
					(degrees <= 90) ? x0++ : x0--; 
				}			    
				if (e2 < dx)
				{
					err = err + dx;
					y0++;
				}
			}
			result.push_back(line);
			line.clear();
		}
	}
	else
	{
		for (int x = -std::max(Nx,Ny); x < std::max(Nx,Ny); x++)
		{
			int x0 = x;
			int y0 = (degrees <= 90) ? 0 : Ny-1;
			int err = dx-dy;

			while ((x0 < 2*Nx) && (y0 < 2*Ny) && (y0 >= -Ny))
			{
				if ((x0 < Nx) && (y0 < Ny) && (x0 >= 0) && (y0 >= 0))
				{
					line.push_back(img.GetVoxel(x0, y0, 0));
				}

				int e2 = 2*err;	
				if (e2 > -dy) 
				{
					err = err - dy;
					x0++;
				}
				if (e2 < dx)
				{
					err = err + dx;
					(degrees <= 90) ? y0++ : y0--; 					
				}
			}
			result.push_back(line);
			line.clear();
		}
	}
	return result;
}


vector<float> filterZeroValuesFromLine(vector<float> line)
{
	vector<float> result; 
	bool previousValueWasZero = false; 

	for (unsigned i = 0; i < line.size(); i++) {
		float pixValue = line[i];

		if (pixValue < 0.01) {
			if (previousValueWasZero) {
				continue;
			} else {
				result.push_back(pixValue);
			}
			previousValueWasZero = true;
		} else {
			result.push_back(pixValue);
			previousValueWasZero = false;
		}
	}
	return result; 
}


vector<Slope> DivideLineToSlopes(vector<float> line)
{
	vector<vector<float> > result;

	SlopesDivider sd; 

	vector<float> filteredLine = filterZeroValuesFromLine(line);

	if (filteredLine.size() > 1) {
		for (unsigned i = 0; i < filteredLine.size(); i++) {
			sd.PushNextPixel(filteredLine[i]);
		}
		sd.FinishCurrentSlope();
	}

	return sd.GetSlopes();
}


double ComputeSlopeHeight(Slope s)
{	
	return s.GetHeight();
}


double ComputeSlopeLength(Slope s)
{
	totalSlopesLength += s.GetLength(); 
	return s.GetLength();
}


double ComputeNormalizedIntegralBelowSlope(Slope s)
{
	vector<float> values = s.GetValues();
	double sum = 0;
	float minValue = s.GetMinValue();

	for (unsigned i = 0; i < values.size(); i++) {
		sum += (values[i] - minValue); 
	}
	return sum / (s.GetLength() * s.GetHeight());
}


double Compute2ndDerivSignChanges(Slope s)
{
	if (s.GetLength() < 3) return -1;  // to distinguish between 0 changes and undefined (here undef=-1)

	int changes = 0;
	vector<float> values = s.GetValues();
	vector<float> derivatives;
	for (unsigned i = 1; i < values.size()-1; i++) {
		derivatives.push_back(values[i-1] - 2*values[i] + values[i+1]); 
	}	
	for (unsigned i = 0; i < derivatives.size()-1; i++) {
		if (derivatives[i]*derivatives[i+1] < 0) changes++;
	}
	return changes;
}


vector<int> CreateHistogram(vector<double> values, double minValue, double maxValue, int numberOfBins)
{
	vector<int> histogram(numberOfBins); 

	double histogramStep = (maxValue - minValue) / numberOfBins; 

	for (unsigned i = 0; i < values.size(); i++) {
		double currentVal = std::min(values[i], maxValue);

		int index = (int)floor((currentVal - minValue) / histogramStep);
		index = std::min(index, numberOfBins - 1);
		index = std::max(index, 0); 

		histogram[index]++; 
	}
	return histogram; 
}


template <class T> void ComputeRSurf(Image3d<T> &img, vector<double> &final_vector)
{
	if (SKIP_FACTOR < 1)
	{
		cerr << endl << "Skip factor must be positive!" << endl;
		exit(1);
	}

	int MAX_SIGN_CHANGES = LEN;	

	// *** Resample To Desired Resolution ***
	Vector3d<float> res = img.GetResolution().GetRes();

	if (img.GetResolution().IsDefined() && ((res.x != res.y) || (res.x != res.z))) {
		float max_res = std::max(std::max(res.x, res.y), res.z);
		Vector3d<float> new_res = max_res;
		ResampleToDesiredResolution(img, new_res, LANCZOS);
	}

	// apply Gauss
	if (SIGMA > 0) {
		if (img.GetSizeZ() > 1) {
			Gauss(img, SIGMA, SIGMA, SIGMA, 3.0, false);
		} else {
			Gauss(img, SIGMA, SIGMA, 0.0, 3.0, false);
		}
	}

	for (unsigned z = 0; z<img.GetSizeZ()-1; z++)
	{	
		for (unsigned y = 0; y<img.GetSizeY()-1; y++)
		{
			for (unsigned x = 0; x<img.GetSizeX()-1; x++)
			{
				if (img.GetVoxel(x,y,z) < 1) img.SetVoxel(x,y,z,0);
			}
		}
	}

	#ifdef I3D_DEBUG
		img.SaveImage("gaussian-output.ics");
	#endif

	// (feature) vectors
	vector<int> histogramOfSlopeHeights (LEN);
	vector<int> histogramOfSlopeLengths (LEN);
	vector<int> histogramOfSlopeIntegrals (LEN);
	vector<int> histogramOf2ndDerivSignChanges (LEN);
	vector<double> histogramOfImageAreaStats (5);

	vector<double> computedSlopeHeights;
	vector<double> computedSlopeLengths;
	vector<double> computedIntegrals;
	vector<double> computed2ndDerivSignChanges;
	int allSlopes = 0;

	// to divide angles from command line
	stringstream ss(ANGLES);
	string ANGLE;
	vector<int> angles, angles3D; 

	// get all angles
	while (getline(ss, ANGLE, ','))
	{
		int angle;
		stringstream convert(ANGLE);
		if (!(convert >> angle) || (angle < 0) || (angle > 180))
		{
			cout << endl << "Wrong angle! Input was: " << angle << endl;
			exit(1);
		}	
		angles.push_back(angle);
	}

	// to process another angle string for 3D images
	if (img.GetSizeZ() > 1)
	{
		// to divide angles from command line
		stringstream ss3D(ANGLES3D);
		string ANGLE3D;

		// get all angles
		while (getline(ss3D, ANGLE3D, ','))
		{
			int angle3D;
			stringstream convert3D(ANGLE3D);
			if (!(convert3D >> angle3D) || (angle3D < 0) || (angle3D > 180))
			{
				cout << endl << "Wrong angle(3D)! Input was: " << angle3D << endl;
				exit(1);
			}	
			angles3D.push_back(angle3D);
		}
	}

	// for all angles
	for (unsigned i = 0; i < angles.size(); i++) 
	{		
		vector<vector<float> > lines;

		// getting pixels as Bresenham line with certain angle
		if (img.GetSizeZ() > 1)
		{
			for (unsigned j = 0; j < angles3D.size(); j++)
			{
				double alpha = angles[i] * M_PI / 180;
				double beta = angles3D[j] * M_PI / 180;

				vector<float> currentOrientation (3);
				currentOrientation.at(0) = (float)(sin(alpha)*cos(beta));
				currentOrientation.at(1) = (float)(sin(alpha)*sin(beta));
				currentOrientation.at(2) = (float)cos(alpha);

				if (UsedOrientations.size() == 0) 
				{
					UsedOrientations.push_back(currentOrientation);
					lines = GetLinesFrom3D(img, angles[i], angles3D[j]);

					// for each line
					for (vector<vector<float> >::iterator it = lines.begin(); it != lines.end(); ++it)
					{
						vector<float> *line = &(*it);

						// divide line to slopes
						vector<Slope> slopesForThisLine = DivideLineToSlopes(*line); 

						for (unsigned i = 0; i < slopesForThisLine.size(); i++) 
						{
							Slope s = slopesForThisLine[i];
							computedSlopeHeights.push_back(ComputeSlopeHeight(s));
							computedSlopeLengths.push_back(ComputeSlopeLength(s));
							computedIntegrals.push_back(ComputeNormalizedIntegralBelowSlope(s));
							computed2ndDerivSignChanges.push_back(Compute2ndDerivSignChanges(s));
							allSlopes++;
						}
					}
				}
				else
				{
					int alreadyUsedOrient = 0;
					for (unsigned k = 0; k < UsedOrientations.size(); k++) {
						vector<float> usedOrient = UsedOrientations[k];
						if ((AreSame( currentOrientation[0], usedOrient[0]) && 
							 AreSame( currentOrientation[1], usedOrient[1]) && 
							 AreSame( currentOrientation[2], usedOrient[2])) ||
							(AreSame(-currentOrientation[0], usedOrient[0]) && 
							 AreSame(-currentOrientation[1], usedOrient[1]) && 
							 AreSame(-currentOrientation[2], usedOrient[2])))
						{							
							alreadyUsedOrient++;
						}
					}

					if (!alreadyUsedOrient) 
					{
						UsedOrientations.push_back(currentOrientation);
						lines = GetLinesFrom3D(img, angles[i], angles3D[j]);

						// for each line
						for (vector<vector<float> >::iterator it = lines.begin(); it != lines.end(); ++it)
						{
							vector<float> *line = &(*it);

							// divide line to slopes
							vector<Slope> slopesForThisLine = DivideLineToSlopes(*line); 

							for (unsigned i = 0; i < slopesForThisLine.size(); i++) 
							{
								Slope s = slopesForThisLine[i];
								computedSlopeHeights.push_back(ComputeSlopeHeight(s));
								computedSlopeLengths.push_back(ComputeSlopeLength(s));
								computedIntegrals.push_back(ComputeNormalizedIntegralBelowSlope(s));
								computed2ndDerivSignChanges.push_back(Compute2ndDerivSignChanges(s));
								allSlopes++;
							}
						}
					}					
				}				
			}
		}
		else
		{			
			lines = GetLines(img, angles[i]);

			// for each line
			for (vector<vector<float> >::iterator it = lines.begin(); it != lines.end(); ++it)
			{
				vector<float> *line = &(*it);

				// divide line to slopes
				vector<Slope> slopesForThisLine = DivideLineToSlopes(*line); 

				for (unsigned i = 0; i < slopesForThisLine.size(); i++) 
				{
					Slope s = slopesForThisLine[i];
					computedSlopeHeights.push_back(ComputeSlopeHeight(s));
					computedSlopeLengths.push_back(ComputeSlopeLength(s));
					computedIntegrals.push_back(ComputeNormalizedIntegralBelowSlope(s));
					computed2ndDerivSignChanges.push_back(Compute2ndDerivSignChanges(s));
					allSlopes++;
				}
			}
		}		
	}

	// compute histogram for slope heights		
	vector<int> histogram = CreateHistogram(computedSlopeHeights, 0, MAX_HEIGHT, LEN);
	histogramOfSlopeHeights.clear();
	for (unsigned i = 0; i < histogram.size(); i++) 
	{
		histogramOfSlopeHeights.push_back(histogram[i]);
	}

	// compute histogram for slope lengths
	histogram.clear();
	histogram = CreateHistogram(computedSlopeLengths, 0, MAX_LENGTH, LEN);
	histogramOfSlopeLengths.clear();
	for (unsigned i = 0; i < histogram.size(); i++) 
	{
		histogramOfSlopeLengths.push_back(histogram[i]);
	}

	// compute histogram for slope integrals
	histogram.clear();
	histogram = CreateHistogram(computedIntegrals, 0, 1, LEN);
	histogramOfSlopeIntegrals.clear();
	for (unsigned i = 0; i < histogram.size(); i++) 
	{
		histogramOfSlopeIntegrals.push_back(histogram[i]);
	}

	// compute histogram for 2nd derivative sign changes
	histogram.clear();
	histogram = CreateHistogram(computed2ndDerivSignChanges, 0, MAX_SIGN_CHANGES, LEN);
	histogramOf2ndDerivSignChanges.clear();
	for (unsigned i = 0; i < histogram.size(); i++) 
	{
		histogramOf2ndDerivSignChanges.push_back(histogram[i]);
	}

	// normalization
	vector<double> normalizedHeight (histogramOfSlopeHeights.size());
	vector<double> normalizedLength (histogramOfSlopeLengths.size());
	vector<double> normalizedIntegral (histogramOfSlopeIntegrals.size());
	vector<double> normalized2ndDerivSignChanges (histogramOf2ndDerivSignChanges.size());

	double norm_height = 0;
	double norm_length = 0;
	double normIntegral = 0;
	double norm2ndDerivSignChanges = 0;

	for (unsigned i = 0; i < histogramOfSlopeHeights.size(); i++) {
		norm_height += histogramOfSlopeHeights[i];
		norm_length += histogramOfSlopeLengths[i];
		normIntegral += histogramOfSlopeIntegrals[i];
		norm2ndDerivSignChanges += histogramOf2ndDerivSignChanges[i];
	}

	for (unsigned i = 0; i < histogramOfSlopeHeights.size(); i++) {
		normalizedHeight[i] = histogramOfSlopeHeights[i] / norm_height;
		normalizedLength[i] = histogramOfSlopeLengths[i] / norm_length;
		normalizedIntegral[i] = histogramOfSlopeIntegrals[i] / normIntegral;
		normalized2ndDerivSignChanges[i] = histogramOf2ndDerivSignChanges[i] / norm2ndDerivSignChanges;
	}

	histogramOfImageAreaStats[0] = 0;
	for (unsigned z = 0; z<img.GetSizeZ()-1; z++)
	{	
		for (unsigned y = 0; y<img.GetSizeY()-1; y++)
		{
			for (unsigned x = 0; x<img.GetSizeX()-1; x++)
			{
				if (img.GetVoxel(x,y,z) < 1) histogramOfImageAreaStats[0]++;
			}
		}
	}
	histogramOfImageAreaStats[1] = histogramOfImageAreaStats[0] / (float)img.GetImageSize();
	histogramOfImageAreaStats[2] = totalSlopesLength;
	histogramOfImageAreaStats[3] = ((totalSlopesLength / (float)angles.size()) * SKIP_FACTOR) / (float)img.GetImageSize();
	histogramOfImageAreaStats[4] = allSlopes;

	for (unsigned i = 0; i < normalizedHeight.size()-1; i++) {
		final_vector.push_back(normalizedHeight[i]); 
	}
	final_vector.push_back(normalizedHeight[normalizedHeight.size()-1]);

	for (unsigned i = 0; i < normalizedLength.size()-1; i++) {
		final_vector.push_back(normalizedLength[i]); 
	}
	final_vector.push_back(normalizedLength[normalizedLength.size()-1]);

	for (unsigned i = 0; i < normalizedIntegral.size()-1; i++) {
		final_vector.push_back(normalizedIntegral[i]); 
	}
	final_vector.push_back(normalizedIntegral[normalizedIntegral.size()-1]);

	for (unsigned i = 0; i < normalized2ndDerivSignChanges.size()-1; i++) {
		final_vector.push_back(normalized2ndDerivSignChanges[i]); 
	}
	final_vector.push_back(normalized2ndDerivSignChanges[normalized2ndDerivSignChanges.size()-1]);

	for (unsigned i = 0; i < histogramOfImageAreaStats.size()-1; i++) {
		final_vector.push_back(histogramOfImageAreaStats[i]); 
	}
	final_vector.push_back(histogramOfImageAreaStats[histogramOfImageAreaStats.size()-1]);
}

//-----------------------------------------------------------------------
// explicit instantiation
//-----------------------------------------------------------------------

template I3D_ALGO_EXPORT void ComputeRSurf<GRAY8>(Image3d<GRAY8> &, vector<double> &);
template I3D_ALGO_EXPORT void ComputeRSurf<GRAY16>(Image3d<GRAY16> &, vector<double> &);

}
