/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * diffusion_filters.cc
 *
 * Diffusion filters
 *
 * Martin Maka (xmaska@fi.muni.cz) 2008 - 2011
 *
 */

#include "diffusion_filters.h"
#include "filters.h"

#ifdef WITH_LAPACK
	// LAPACK function for calculating eigenvalues and eigenvectors (double precision)
    extern "C" void dsyev_(const char &jobz, const char &uplo, const int &n, double *a, const int &lda, double *w, double *work, const int &lwork, int &info);
#endif // WITH_LAPACK

namespace i3d
{
	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                         Local Functions                     //
	//                                                             //
	/////////////////////////////////////////////////////////////////

	/**********************************************************************/
	
	/** Solve a tridiagonal system of linear equations using the Thomas algorithm. */
	template <class T> static void TridiagonalMatrixSolver(T *a, T *b, T *c, T *d, T *x, size_t n)
	{
		for (size_t i = 0; i < n - 1; ++i)
		{
			c[i] /= a[i];
			a[i + 1] -= c[i] * b[i]; 
		}

		for (size_t i = 1; i < n; ++i)
		{
			d[i] -= c[i - 1] * d[i - 1];
		}

		x[n - 1] = d[n - 1] / a[n - 1];

		for (int i = static_cast<int>(n - 2); i > -1; --i)
		{
			x[i] = (d[i] - b[i] * x[i + 1]) / a[i];
		}
	}

	/**********************************************************************/

	/** Forward and backward substitution of the Thomas algorithm intended for a linear diffusion filtering. */
	template <class VOXEL> static void ForwardBackwardSubstitution(const VOXEL *M, const VOXEL *L, VOXEL *d, size_t n, VOXEL tau, size_t shift)
	{
		size_t tmp = 0;

		for (size_t i = 1; i < n; ++i, tmp += shift)
		{
			d[tmp + shift] -= L[i - 1] * d[tmp];
		}

		d[tmp] /= M[n - 1];
		tmp -= shift;

		for (int i = static_cast<int>(n - 2); i > -1; --i, tmp -= shift)
		{
			d[tmp] += tau * d[tmp + shift];
			d[tmp] /= M[i];
		}
	}

	/**********************************************************************/

	/** Prepare the LR decomposition of a tridiagonal matrix intended for a linear diffusion filtering. */
	template <class VOXEL> static void PrepareLRDecomposition(VOXEL *M, VOXEL *L, size_t n, VOXEL tau)
	{
		M[0] = 1 + tau;

		for (size_t i = 0; i < n - 1; ++i)
		{
			L[i] = -tau / M[i];
			M[i + 1] = M[0] + tau + tau * L[i];
		}

		M[n - 1] -= tau;
	}

	/**********************************************************************/

	/** Create a tridiagonal matrix intended for a AOS scheme. */
	template <class T> static void PrepareDiagonals(T *Diag, T *UDiag, T *LDiag, const T *data, size_t N, T tau, size_t shift)
	{
		UDiag[0] = -tau * (*data + *(data + shift));
		LDiag[0] = UDiag[0];
		Diag[0] = 1 - UDiag[0];
		data += shift;

		for (size_t i = 1; i < N - 1; ++i, data += shift)
		{
			UDiag[i] = -tau * (*data + *(data + shift));
			LDiag[i] = UDiag[i];
			Diag[i] = 1 - UDiag[i] - UDiag[i - 1];
		}

		Diag[N - 1] = 1 - UDiag[N - 2];
	}

	/**********************************************************************/

	/** Copy values from one buffer to the other one. */
	template <class T> static void CopyFrom(const T *src, T *dst, size_t num, size_t step)
	{
		for (size_t i= 0; i < num; ++i, src += step, ++dst)
		{
			*dst = *src;
		}
	}

	/**********************************************************************/

	/** Add values from one buffer to the other one. */
	template <class T> static void Add(const T *src, T *dst, size_t num, size_t step)
	{
		for (size_t i = 0; i < num; ++i, ++src, dst += step)
		{
			*dst += *src;
		}
	}

	/**********************************************************************/

	/** Add values from one buffer to the other one and multiply them by a given coefficient. */
	template <class T> static void AddAndAvg(const T *src, T *dst, size_t num, size_t step, T coef)
	{
		for (size_t i = 0; i < num; ++i, ++src, dst += step)
		{
			*dst += *src;
			*dst *= coef;
		}
	}
 
	/**********************************************************************/

	/** Compute the eigenvalue decomposition of a symmetric 2x2 matrix in the form A = [a b; b c]. The eigenvalues are always sorted in the ascending order. */
	template <class T> static void EigDecomposition(T a, T b, T c, double *eigval, double *eigvec)
	{
		if (b == T(0) && (c == T(0) || a == T(0)))
		{
			if (c == T(0))
			{
				eigval[0] = 0.0;
				eigval[1] = a;
				eigvec[0] = 0.0;
				eigvec[1] = 1.0;
				eigvec[2] = 1.0;
				eigvec[3] = 0.0;
			}
			else
			{
				eigval[0] = 0.0;
				eigval[1] = c;
				eigvec[0] = 1.0;
				eigvec[1] = 0.0;
				eigvec[2] = 0.0;
				eigvec[3] = 1.0;
			}
		}
		else
		{
			T disc = std::sqrt((a - c) * (a - c) + 4 * b * b);

			eigval[0] = (a + c - disc) / 2;
			eigval[1] = (a + c + disc) / 2;

			eigvec[0] = -b;  eigvec[1] = a - eigval[0];
			eigvec[2] = -b;  eigvec[3] = a - eigval[1];

			double v1mag = std::sqrt(eigvec[0] * eigvec[0] + eigvec[1] * eigvec[1]);  
			double v2mag = std::sqrt(eigvec[2] * eigvec[2] + eigvec[3] * eigvec[3]);  

			if (v1mag >= std::numeric_limits<double>::epsilon()) 
			{ 
				eigvec[0] /= v1mag;  
				eigvec[1] /= v1mag;
			}

			if (v2mag >= std::numeric_limits<double>::epsilon()) 
			{
				eigvec[2] /= v2mag;  
				eigvec[3] /= v2mag;
			}
		}
	}

	/**********************************************************************/

	/** Compute the eigenvalue decomposition of a symmetric 3x3 matrix in the form A = [Da Db Dc; Db Dd De; Dc De Df]. The eigenvalues are always sorted in the ascending order. */
	template <class T> static void EigDecomposition3D(T a, T b, T c, T d, T e, T f, double *eigval, double *eigvec)
	{
		#ifdef WITH_LAPACK

			double tmp[102];
			int info;
		
			// create a 3x3 symmetric matrix 
			eigvec[0] = a;
			eigvec[3] = b;
			eigvec[4] = d;
			eigvec[6] = c;
			eigvec[7] = e;
			eigvec[8] = f;
		
			// compute its eigenvalues and eigenvectors
			dsyev_('V', 'U', 3, eigvec, 3, eigval, tmp, 102, info);

		#else   // WITH_LAPACK

			throw InternalException("LAPACK library required!");

		#endif  // WITH_LAPACK
	}

	/**********************************************************************/

	/** Construct a structure tensor J = [Da Db; Db Dc] for each image pixel. */ 
	template <class T> static void StructureTensor(const Image3d<T> &in, Image3d<T> &Da, Image3d<T> &Db, Image3d<T> &Dc)
	{
		// the image size
		size_t sx = in.GetSizeX();
		size_t sy = in.GetSizeY();

		// the image resolution
		Vector3d<float> res = in.GetResolution().GetRes();
		
		T coef_x = res.x * res.x / 4;
		T coef_y = res.y * res.y / 4;
		T coef_xy = res.x * res.y / 4;

		const T *src = in.begin() + 1;
		T *a = Da.begin();
		T *b = Db.begin();
		T *c = Dc.begin();
		
		// upper-left corner
		*a = 0; *b = 0; *c = 0; ++a; ++b; ++c;

		// the first row without both corners
		for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++src)
		{
			*a = *(src + 1) - *(src - 1); 
			*a *= coef_x * (*a); 
			*b = 0; 
			*c = 0;
		}

		// upper-right corner
		*a = 0; *b = 0; *c = 0; ++a; ++b; ++c; ++src;

		// the remaining rows except the last one
		for (size_t y = 2; y < sy; ++y)
		{
			// the first pixel in the row
			*a = 0; 
			*b = 0; 
			*c = *(src + sx) - *(src - sx); 
			*c *= coef_y * (*c);
			++a; ++b; ++c; ++src;	

			// the remaining pixels except the last one
			for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++src)
			{
				*a = *(src + 1) - *(src - 1);
				*c = *(src + sx) - *(src - sx);
				*b = coef_xy * (*a) * (*c);
				*a *= coef_x * (*a);
				*c *= coef_y * (*c);
			}
			
			// the last pixel in the row
			*a = 0; 
			*b = 0; 
			*c = *(src + sx) - *(src - sx); 
			*c *= *c * coef_y;
			++a; ++b; ++c; ++src;
		}

		// lower-left corner
		*a = 0; *b = 0; *c = 0; ++a; ++b; ++c; ++src;

		// the last row without both corners
		for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++src)
		{
			*a = *(src + 1) - *(src - 1); 
			*a *= coef_x * (*a); 
			*b = 0; 
			*c = 0;
		}

		// lower-right corner
		*a = 0; *b = 0; *c = 0;
	}

	/**********************************************************************/

	/** Construct a structure tensor J = [Da Db Dc; Db Dd De; Dc De Df] for each image voxel. */ 
	template <class T> static void StructureTensor3D(const Image3d<T> &in, Image3d<T> &Da, Image3d<T> &Db, Image3d<T> &Dc, Image3d<T> &Dd, Image3d<T> &De, Image3d<T> &Df)
	{
		// the image size
		size_t sx = in.GetSizeX();
		size_t sy = in.GetSizeY();
		size_t sz = in.GetSizeZ();

		// the size of one slice
		size_t slice = sx * sy;

		// the image resolution
		Vector3d<float> res = in.GetResolution().GetRes();
		
		T coef_xx = res.x * res.x / 4;
		T coef_yy = res.y * res.y / 4;
		T coef_zz = res.z * res.z / 4;
		T coef_xy = res.x * res.y / 4;
		T coef_xz = res.x * res.z / 4;
		T coef_yz = res.y * res.z / 4;

		const T *src = in.begin() + 1;
		T *a = Da.begin();
		T *b = Db.begin();
		T *c = Dc.begin();
		T *d = Dd.begin();
		T *e = De.begin();
		T *f = Df.begin();
		
		// process the first slice

		// upper-left corner
		*a = 0; *b = 0; *c = 0; *d = 0; *e = 0; *f = 0; ++a; ++b; ++c; ++d; ++e; ++f;

		// the first row without both corners
		for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++d, ++e, ++f, ++src)
		{
			*a = *(src + 1) - *(src - 1); *a *= coef_xx * (*a); 
			*b = 0; 
			*c = 0;
			*d = 0;
			*e = 0;
			*f = 0;
		}

		// upper-right corner
		*a = 0; *b = 0; *c = 0; *d = 0; *e = 0; *f = 0; ++a; ++b; ++c; ++d; ++e; ++f; ++src;

		// the remaining rows except the last one
		for (size_t y = 2; y < sy; ++y)
		{
			// the first pixel in the row
			*a = 0; 
			*b = 0;
			*c = 0;
			*d = *(src + sx) - *(src - sx); *d *= coef_yy * (*d);
			*e = 0;
			*f = 0;
			++a; ++b; ++c; ++d; ++e; ++f; ++src;	

			// the remaining pixels except the last one
			for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++d, ++e, ++f, ++src)
			{
				*a = *(src + 1) - *(src - 1);
				*d = *(src + sx) - *(src - sx);
				*b = coef_xy * (*a) * (*d);
				*a *= coef_xx * (*a);
				*d *= coef_yy * (*d);
				*c = 0;
				*e = 0;
				*f = 0;
			}
			
			// the last pixel in the row
			*a = 0; 
			*b = 0; 
			*c = 0;
			*d = *(src + sx) - *(src - sx); *d *= coef_yy * (*d);
			*e = 0;
			*f = 0;
			++a; ++b; ++c; ++d; ++e; ++f; ++src;
		}

		// lower-left corner
		*a = 0; *b = 0; *c = 0; *d = 0; *e = 0; *f = 0; ++a; ++b; ++c; ++d; ++e; ++f; ++src;

		// the last row without both corners
		for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++d, ++e, ++f, ++src)
		{
			*a = *(src + 1) - *(src - 1); *a *= coef_xx * (*a);
			*b = 0;
			*c = 0;
			*d = 0;
			*e = 0;
			*f = 0;
		}

		// lower-right corner
		*a = 0; *b = 0; *c = 0; *d = 0; *e = 0; *f = 0; ++a; ++b; ++c; ++d; ++e; ++f; ++src;

		// process the remaining slices except the last one
		for (size_t z = 2; z < sz; ++z)
		{
			// upper-left corner
			*a = 0; 
			*b = 0; 
			*c = 0;
			*d = 0;
			*e = 0;
			*f = *(src + slice) - *(src - slice); *f *= coef_zz * (*f);
			++a; ++b; ++c; ++d; ++e; ++f; ++src;

			// the first row without both corners
			for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++d, ++e, ++f, ++src)
			{
				*a = *(src + 1) - *(src - 1); 
				*f = *(src + slice) - *(src - slice); 
				*c = coef_xz * (*a) * (*f);
				*a *= coef_xx * (*a); 
				*f *= coef_zz * (*f);
				*b = 0; 
				*d = 0;
				*e = 0;
			}

			// upper-right corner
			*a = 0; 
			*b = 0; 
			*c = 0;
			*d = 0;
			*e = 0;
			*f = *(src + slice) - *(src - slice); *f *= coef_zz * (*f);
			++a; ++b; ++c; ++d; ++e; ++f; ++src;

			// the remaining rows except the last one
			for (size_t y = 2; y < sy; ++y)
			{
				// the first pixel in the row
				*a = 0; 
				*b = 0;
				*c = 0;
				*d = *(src + sx) - *(src - sx); 
				*f = *(src + slice) - *(src - slice);
				*e = coef_yz * (*d) * (*f);
				*d *= coef_yy * (*d);
				*f *= coef_zz * (*f);
				++a; ++b; ++c; ++d; ++e; ++f; ++src;	

				// the remaining pixels except the last one
				for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++d, ++e, ++f, ++src)
				{
					*a = *(src + 1) - *(src - 1);
					*d = *(src + sx) - *(src - sx);
					*f = *(src + slice) - *(src - slice);
					*b = coef_xy * (*a) * (*d);
					*c = coef_xz * (*a) * (*f);
					*e = coef_yz * (*d) * (*f);
					*a *= coef_xx * (*a);
					*d *= coef_yy * (*d);
					*f *= coef_zz * (*f);
				}

				// the last pixel in the row
				*a = 0; 
				*b = 0;
				*c = 0;
				*d = *(src + sx) - *(src - sx); 
				*f = *(src + slice) - *(src - slice);
				*e = coef_yz * (*d) * (*f);
				*d *= coef_yy * (*d);
				*f *= coef_zz * (*f);
				++a; ++b; ++c; ++d; ++e; ++f; ++src;	
			}

			// lower-left corner
			*a = 0; 
			*b = 0; 
			*c = 0;
			*d = 0;
			*e = 0;
			*f = *(src + slice) - *(src - slice); *f *= coef_zz * (*f);
			++a; ++b; ++c; ++d; ++e; ++f; ++src;

			// the last row without both corners
			for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++d, ++e, ++f, ++src)
			{
				*a = *(src + 1) - *(src - 1); 
				*f = *(src + slice) - *(src - slice); 
				*c = coef_xz * (*a) * (*f);
				*a *= coef_xx * (*a); 
				*f *= coef_zz * (*f);
				*b = 0; 
				*d = 0;
				*e = 0;
			}

			// lower-right corner
			*a = 0; 
			*b = 0; 
			*c = 0;
			*d = 0;
			*e = 0;
			*f = *(src + slice) - *(src - slice); *f *= coef_zz * (*f);
			++a; ++b; ++c; ++d; ++e; ++f; ++src;
		}

		// process the last slice

		// upper-left corner
		*a = 0; *b = 0; *c = 0; *d = 0; *e = 0; *f = 0; ++a; ++b; ++c; ++d; ++e; ++f; ++src;

		// the first row without both corners
		for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++d, ++e, ++f, ++src)
		{
			*a = *(src + 1) - *(src - 1); *a *= coef_xx * (*a); 
			*b = 0; 
			*c = 0;
			*d = 0;
			*e = 0;
			*f = 0;
		}

		// upper-right corner
		*a = 0; *b = 0; *c = 0; *d = 0; *e = 0; *f = 0; ++a; ++b; ++c; ++d; ++e; ++f; ++src;

		// the remaining rows except the last one
		for (size_t y = 2; y < sy; ++y)
		{
			// the first pixel in the row
			*a = 0; 
			*b = 0;
			*c = 0;
			*d = *(src + sx) - *(src - sx); *d *= coef_yy * (*d);
			*e = 0;
			*f = 0;
			++a; ++b; ++c; ++d; ++e; ++f; ++src;	

			// the remaining pixels except the last one
			for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++d, ++e, ++f, ++src)
			{
				*a = *(src + 1) - *(src - 1);
				*d = *(src + sx) - *(src - sx);
				*b = coef_xy * (*a) * (*d);
				*a *= coef_xx * (*a);
				*d *= coef_yy * (*d);
				*c = 0;
				*e = 0;
				*f = 0;
			}
			
			// the last pixel in the row
			*a = 0; 
			*b = 0; 
			*c = 0;
			*d = *(src + sx) - *(src - sx); *d *= *d * coef_yy;
			*e = 0;
			*f = 0;
			++a; ++b; ++c; ++d; ++e; ++f; ++src;
		}

		// lower-left corner
		*a = 0; *b = 0; *c = 0; *d = 0; *e = 0; *f = 0; ++a; ++b; ++c; ++d; ++e; ++f; ++src;

		// the last row without both corners
		for (size_t x = 2; x < sx; ++x, ++a, ++b, ++c, ++d, ++e, ++f, ++src)
		{
			*a = *(src + 1) - *(src - 1); *a *= coef_xx * (*a);
			*b = 0;
			*c = 0;
			*d = 0;
			*e = 0;
			*f = 0;
		}

		// lower-right corner
		*a = 0; *b = 0; *c = 0; *d = 0; *e = 0; *f = 0;
	}
			
	/**********************************************************************/

	/** Construct a diffusion tensor D = [Da Db; Db Dc] for the edge-enhancing diffusion. */
	template <class T> static void EEDTensor(Image3d<T> &Da, Image3d<T> &Db, Image3d<T> &Dc, T lambda)
	{
		//HELP remove eigenvalue decomposition
		double eigval[2] = {0}, eigvec[4] = {0};

		T *a = Da.begin();
		T *b = Db.begin();
		T *c = Dc.begin();
		T *last = Da.end();

		while (a != last)
		{
			// compute eigenvalues and eigenvectors of a structure tensor
			// the eigenvalues are always sorted in the ascending order
			EigDecomposition(*a, *b, *c, eigval, eigvec);

			// transform eigenvalues
			eigval[0] = 1.0;
			eigval[1] = (eigval[1] <= std::numeric_limits<double>::epsilon()) ? 1.0 : 1.0 - std::exp(-3.31488 / std::pow(eigval[1] / lambda, 4));

			// construct the diffusion tensor that has the same eigenvectors as the structure tensor and transformed eigenvalues
			*a = static_cast<T>(eigval[0] * eigvec[0] * eigvec[0] + eigval[1] * eigvec[2] * eigvec[2]);
			*b = static_cast<T>(eigval[0] * eigvec[0] * eigvec[1] + eigval[1] * eigvec[2] * eigvec[3]);
			*c = static_cast<T>(eigval[0] * eigvec[1] * eigvec[1] + eigval[1] * eigvec[3] * eigvec[3]);

			++a; ++b; ++c;
		}
	}

	/**********************************************************************/

	/** Construct a diffusion tensor D = [Da Db Dc; Db Dd De; Dc De Df] for the edge-enhancing diffusion. */
	template <class T> static void EEDTensor3D(Image3d<T> &Da, Image3d<T> &Db, Image3d<T> &Dc, Image3d<T> &Dd, Image3d<T> &De, Image3d<T> &Df, T lambda)
	{
		//HELP remove eigenvalue decomposition 
		double eigval[3] = {0}, eigvec[9] = {0};

		T *a = Da.begin();
		T *b = Db.begin();
		T *c = Dc.begin();
		T *d = Dd.begin();
		T *e = De.begin();
		T *f = Df.begin();
		T *last = Da.end();

		while (a != last)
		{
			// compute eigenvalues and eigenvectors of a structure tensor
			// the eigenvalues are always sorted in the ascending order
			EigDecomposition3D(*a, *b, *c, *d, *e, *f, eigval, eigvec);

			// transform eignevalues
			eigval[0] = 1.0;
			eigval[1] = 1.0;
			eigval[2] = (eigval[2] <= std::numeric_limits<double>::epsilon()) ? 1.0 : 1.0 - std::exp(-3.31488 / std::pow(eigval[2] / lambda, 4));

			// construct the diffusion tensor that has the same eigenvectors as the structure tensor and transformed eigenvalues
			*a = static_cast<T>(eigval[0] * eigvec[0] * eigvec[0] + eigval[1] * eigvec[3] * eigvec[3] + eigval[2] * eigvec[6] * eigvec[6]);
			*b = static_cast<T>(eigval[0] * eigvec[0] * eigvec[1] + eigval[1] * eigvec[3] * eigvec[4] + eigval[2] * eigvec[6] * eigvec[7]);
			*c = static_cast<T>(eigval[0] * eigvec[0] * eigvec[2] + eigval[1] * eigvec[3] * eigvec[5] + eigval[2] * eigvec[6] * eigvec[8]);
			*d = static_cast<T>(eigval[0] * eigvec[1] * eigvec[1] + eigval[1] * eigvec[4] * eigvec[4] + eigval[2] * eigvec[7] * eigvec[7]);
			*e = static_cast<T>(eigval[0] * eigvec[1] * eigvec[2] + eigval[1] * eigvec[4] * eigvec[5] + eigval[2] * eigvec[7] * eigvec[8]);
			*f = static_cast<T>(eigval[0] * eigvec[2] * eigvec[2] + eigval[1] * eigvec[5] * eigvec[5] + eigval[2] * eigvec[8] * eigvec[8]);

			++a; ++b; ++c; ++d; ++e; ++f;
		}
	}

	/**********************************************************************/
	
	/** Construct a diffusion tensor D = [Da Db; Db Dc] for the coherence-enhancing diffusion. */
	template <class T> static void CEDTensor(Image3d<T> &Da, Image3d<T> &Db, Image3d<T> &Dc)
	{
		double eigval[2] = {0}, eigvec[4] = {0}, coherence;

		T *a = Da.begin();
		T *b = Db.begin();
		T *c = Dc.begin();
		T *last = Da.end();

		while (a != last)
		{
			// compute eigenvalues and eigenvectors of a structure tensor
			// the eigenvalues are always sorted in the ascending order
			EigDecomposition(*a, *b, *c, eigval, eigvec);

			// compute the coherence
			coherence = eigval[0] - eigval[1];
			coherence *= coherence;

			// transform eigenvalues
			eigval[0] = (coherence < std::numeric_limits<double>::epsilon()) ? 0.01 : 0.01 + 0.99 * std::exp(-3.31488 / coherence);
			eigval[1] = 0.01;

			// construct the diffusion tensor that has the same eigenvectors as the structure tensor and transformed eigenvalues
			*a = static_cast<T>(eigval[0] * eigvec[0] * eigvec[0] + eigval[1] * eigvec[2] * eigvec[2]);
			*b = static_cast<T>(eigval[0] * eigvec[0] * eigvec[1] + eigval[1] * eigvec[2] * eigvec[3]);
			*c = static_cast<T>(eigval[0] * eigvec[1] * eigvec[1] + eigval[1] * eigvec[3] * eigvec[3]);

			++a; ++b; ++c;
		}
	}

	/**********************************************************************/
	
	/** Construct a diffusion tensor D = [Da Db Dc; Db Dd De; Dc De Df] for the coherence-enhancing diffusion. */
	template <class T> static void CEDTensor3D(Image3d<T> &Da, Image3d<T> &Db, Image3d<T> &Dc, Image3d<T> &Dd, Image3d<T> &De, Image3d<T> &Df)
	{
		double eigval[3] = {0}, eigvec[9] = {0}, coherence;

		T *a = Da.begin();
		T *b = Db.begin();
		T *c = Dc.begin();
		T *d = Dd.begin();
		T *e = De.begin();
		T *f = Df.begin();
		T *last = Da.end();

		while (a != last)
		{
			// compute eigenvalues and eigenvectors of a structure tensor
			// the eigenvalues are always sorted in the ascending order
			EigDecomposition3D(*a, *b, *c, *d, *e, *f, eigval, eigvec);

			// compute the coherence
			coherence = (eigval[2] - eigval[1]) * (eigval[2] - eigval[1]) + (eigval[2] - eigval[0]) * (eigval[2] - eigval[0]) + (eigval[1] - eigval[0]) * (eigval[1] - eigval[0]);
						
			// transform eigenvalues
			eigval[0] = (coherence < std::numeric_limits<double>::epsilon()) ? 0.01 : 0.01 + 0.99 * std::exp(-3.31488 / coherence);
			eigval[1] = 0.01;
			eigval[2] = 0.01;

			// construct the diffusion tensor that has the same eigenvectors as the structure tensor and transformed eigenvalues
			*a = static_cast<T>(eigval[0] * eigvec[0] * eigvec[0] + eigval[1] * eigvec[3] * eigvec[3] + eigval[2] * eigvec[6] * eigvec[6]);
			*b = static_cast<T>(eigval[0] * eigvec[0] * eigvec[1] + eigval[1] * eigvec[3] * eigvec[4] + eigval[2] * eigvec[6] * eigvec[7]);
			*c = static_cast<T>(eigval[0] * eigvec[0] * eigvec[2] + eigval[1] * eigvec[3] * eigvec[5] + eigval[2] * eigvec[6] * eigvec[8]);
			*d = static_cast<T>(eigval[0] * eigvec[1] * eigvec[1] + eigval[1] * eigvec[4] * eigvec[4] + eigval[2] * eigvec[7] * eigvec[7]);
			*e = static_cast<T>(eigval[0] * eigvec[1] * eigvec[2] + eigval[1] * eigvec[4] * eigvec[5] + eigval[2] * eigvec[7] * eigvec[8]);
			*f = static_cast<T>(eigval[0] * eigvec[2] * eigvec[2] + eigval[1] * eigvec[5] * eigvec[5] + eigval[2] * eigvec[8] * eigvec[8]);

			++a; ++b; ++c; ++d; ++e; ++f;
		}
	}

	/**********************************************************************/

	/** Execute one step of the anisotropic diffusion for given time step tau and diffusion tensor [Da Db; Db Dc] using the semi-implicit AOS-stabilized scheme. */
	template <class T> static void AnisoDiffStep_AOS(const Image3d<T> &in, const Image3d<T> &Da, const Image3d<T> &Db, const Image3d<T> &Dc, Image3d<T> &out, T tau)
	{
		size_t sx = in.GetSizeX();
		size_t sy = in.GetSizeY();
		size_t N = std::max(sx, sy);
		T *r = new T[N];
		T *Diag = new T[N];
		T *UDiag = new T[N - 1];
		T *LDiag = new T[N - 1];
		T *sol = new T[N];

		const T *u = in.begin();
		const T *b = Db.begin() + sx;

		// create the image containing the right-hand side for the AOS scheme
		i3d::Image3d<T> tmp;
		tmp.CopyMetaData(in);
		T *tmp_ptr = tmp.begin();

		// process the first row
		memcpy(tmp_ptr, u, sx * sizeof(T));
		u += sx; tmp_ptr += sx;

		// process the remaining rows except the last one
		for (size_t y = 1; y < sy - 1; ++y, ++u, ++tmp_ptr, ++b)
		{
			*tmp_ptr = *u; 
			++u; ++tmp_ptr; ++b;

			for (size_t x = 1; x < sx - 1; ++x, ++u, ++tmp_ptr, ++b)
			{
				*tmp_ptr = *u + T(0.25) * tau * (*(u - sx - 1) * (*(b - 1) + *(b - sx)) -
					                             *(u - sx + 1) * (*(b + 1) + *(b - sx)) -	 
												 *(u + sx - 1) * (*(b - 1) + *(b + sx)) +
												 *(u + sx + 1) * (*(b + 1) + *(b + sx)));
			}

			*tmp_ptr = *u;
		}

		// process the last row
		memcpy(tmp_ptr, u, sx * sizeof(T));

		/*************** scan in the x axis ***************/
		const T *a = Da.begin();
		T *ptr = out.begin();
		tmp_ptr = tmp.begin();

		for (size_t y = 0; y < sy; ++y, a += sx, tmp_ptr += sx, ptr += sx)
		{
			memcpy(r, tmp_ptr, sx * sizeof(T));
			PrepareDiagonals(Diag, UDiag, LDiag, a, sx, tau, 1);
			TridiagonalMatrixSolver(Diag, UDiag, LDiag, r, ptr, sx);
		}

		/*************** scan in the y axis ***************/
		const T *c;

		for (size_t x = 0; x < sx; ++x)
		{
			c = Dc.begin() + x;
			tmp_ptr = tmp.begin() + x;
			ptr = out.begin() + x;

			CopyFrom(tmp_ptr, r, sy, sx);
			PrepareDiagonals(Diag, UDiag, LDiag, c, sy, tau, sx);
			TridiagonalMatrixSolver(Diag, UDiag, LDiag, r, sol, sy);
			AddAndAvg(sol, ptr, sy, sx, T(0.5));
		}

		delete [] r;
		delete [] Diag;
		delete [] UDiag;
		delete [] LDiag;
		delete [] sol;
	}

	/**********************************************************************/

	/** Execute one step of the anisotropic diffusion for given time step tau and diffusion tensor [Da Db Dc; Db Dd De; Dc De Df] using the semi-implicit AOS-stabilized scheme. */
	template <class T> static void AnisoDiffStep3D_AOS(const Image3d<T> &in, const Image3d<T> &Da, const Image3d<T> &Db, const Image3d<T> &Dc, const Image3d<T> &Dd, 
													   const Image3d<T> &De, const Image3d<T> &Df, Image3d<T> &out, T tau)
	{
		size_t sx = in.GetSizeX();
		size_t sy = in.GetSizeY();
		size_t sz = in.GetSizeZ();
		size_t slice = sx * sy;
		size_t N = std::max(std::max(sx, sy), sz);
		T *r = new T[N];
		T *Diag = new T[N];
		T *UDiag = new T[N - 1];
		T *LDiag = new T[N - 1];
		T *sol = new T[N];

		const T *u = in.begin();
		const T *b = Db.begin() + sx;
		const T *c = Dc.begin() + slice + 1;
		const T *e = De.begin() + slice + sx;
		T coef = 3 * tau / 2;
		T avg_coef = static_cast<T>(1.0 / 3.0);

		// create the image containing the right-hand side for the AOS scheme
		i3d::Image3d<T> tmp;
		tmp.CopyMetaData(in);
		T *tmp_ptr = tmp.begin();

		/*----- process the first slice -----*/

		// process the first row
		memcpy(tmp_ptr, u, sx * sizeof(T));
		u += sx; tmp_ptr += sx;

		// process the remaining rows except the last one
		for (size_t y = 1; y < sy - 1; ++y, ++u, ++tmp_ptr, ++b)
		{
			*tmp_ptr = *u; 
			++u; ++tmp_ptr; ++b;

			for (size_t x = 1; x < sx - 1; ++x, ++u, ++tmp_ptr, ++b)
			{
				*tmp_ptr = *u + T(0.25) * tau * (*(u - sx - 1) * (*(b - 1) + *(b - sx)) -
					                             *(u - sx + 1) * (*(b + 1) + *(b - sx)) -	 
												 *(u + sx - 1) * (*(b - 1) + *(b + sx)) +
												 *(u + sx + 1) * (*(b + 1) + *(b + sx)));
			}

			*tmp_ptr = *u;
		}

		// process the last row
		memcpy(tmp_ptr, u, sx * sizeof(T));
		u += sx; tmp_ptr += sx; b += 2 * sx + 1;

		/*----- process the remaining slices except the last one -----*/
		for (size_t z = 1; z < sz - 1; ++z, ++u, ++tmp_ptr, b += 2 * sx + 1, c += 2, e += 2 * sx)
		{
			// process the first row
			*tmp_ptr = *u; 
			++u; ++tmp_ptr;

			for (size_t x = 1; x < sx - 1; ++x, ++u, ++tmp_ptr, ++c)
			{
				*tmp_ptr = *u + T(0.25) * tau * (*(u - slice - 1) * (*(c - 1) + *(c - slice)) -
											     *(u - slice + 1) * (*(c + 1) + *(c - slice)) -
											     *(u + slice - 1) * (*(c - 1) + *(c + slice)) +
											     *(u + slice + 1) * (*(c + 1) + *(c + slice)));
			}

			*tmp_ptr = *u;
			++u; ++tmp_ptr; c += 2;

			// process the remaining rows except the last one
			for (size_t y = 1; y < sy - 1; ++y, ++u, ++tmp_ptr, b += 2, c += 2, ++e)
			{
				*tmp_ptr = *u + T(0.25) * tau * (*(u - slice - sx) * (*(e - sx) + *(e - slice)) -
											     *(u - slice + sx) * (*(e + sx) + *(e - slice)) -
											     *(u + slice - sx) * (*(e - sx) + *(e + slice)) +
											     *(u + slice + sx) * (*(e + sx) + *(e + slice)));
				++u; ++tmp_ptr; ++e;

				for (size_t x = 1; x < sx - 1; ++x, ++u, ++tmp_ptr, ++b, ++c, ++e)
				{
					*tmp_ptr = *u + T(0.25) * tau * (*(u - sx - 1) * (*(b - 1) + *(b - sx)) -
						                             *(u - sx + 1) * (*(b + 1) + *(b - sx)) -	 
						                             *(u + sx - 1) * (*(b - 1) + *(b + sx)) +
										             *(u + sx + 1) * (*(b + 1) + *(b + sx)) +
													 
													 *(u - slice - 1) * (*(c - 1) + *(c - slice)) -
													 *(u - slice + 1) * (*(c + 1) + *(c - slice)) -
													 *(u + slice - 1) * (*(c - 1) + *(c + slice)) +
													 *(u + slice + 1) * (*(c + 1) + *(c + slice)) +

													 *(u - slice - sx) * (*(e - sx) + *(e - slice)) -
													 *(u - slice + sx) * (*(e + sx) + *(e - slice)) -
													 *(u + slice - sx) * (*(e - sx) + *(e + slice)) +
													 *(u + slice + sx) * (*(e + sx) + *(e + slice)));
												
				}

				*tmp_ptr = *u + T(0.25) * tau * (*(u - slice - sx) * (*(e - sx) + *(e - slice)) -
						  					     *(u - slice + sx) * (*(e + sx) + *(e - slice)) -
												 *(u + slice - sx) * (*(e - sx) + *(e + slice)) +
												 *(u + slice + sx) * (*(e + sx) + *(e + slice)));
			}

			// process the last row
			*tmp_ptr = *u; 
			++u; ++tmp_ptr;

			for (size_t x = 1; x < sx - 1; ++x, ++u, ++tmp_ptr, ++c)
			{
				*tmp_ptr = *u + T(0.25) * tau * (*(u - slice - 1) * (*(c - 1) + *(c - slice)) -
											     *(u - slice + 1) * (*(c + 1) + *(c - slice)) -
											     *(u + slice - 1) * (*(c - 1) + *(c + slice)) +
											     *(u + slice + 1) * (*(c + 1) + *(c + slice)));
			}

			*tmp_ptr = *u;
		}

		/*----- process the last slice -----*/

		// process the first row
		memcpy(tmp_ptr, u, sx * sizeof(T));
		u += sx; tmp_ptr += sx;

		//HELP vyresit pres increment ne pres end()
		b = Db.end() - slice + sx;
		
		// process the remaining rows except the last one
		for (size_t y = 1; y < sy - 1; ++y, ++u, ++tmp_ptr, ++b)
		{
			*tmp_ptr = *u; 
			++u; ++tmp_ptr; ++b;

			for (size_t x = 1; x < sx - 1; ++x, ++u, ++tmp_ptr, ++b)
			{
				*tmp_ptr = *u + T(0.25) * tau * (*(u - sx - 1) * (*(b - 1) + *(b - sx)) -
					                             *(u - sx + 1) * (*(b + 1) + *(b - sx)) -	 
												 *(u + sx - 1) * (*(b - 1) + *(b + sx)) +
												 *(u + sx + 1) * (*(b + 1) + *(b + sx)));
			}

			*tmp_ptr = *u;
		}

		// process the last row
		memcpy(tmp_ptr, u, sx * sizeof(T));

		/*************** scan in the x axis ***************/
		const T *a = Da.begin();
		T *ptr = out.begin();
		tmp_ptr = tmp.begin();

		for (size_t i = 0; i < sz * sy; ++i, a += sx, tmp_ptr += sx, ptr += sx)
		{
			memcpy(r, tmp_ptr, sx * sizeof(T));
			PrepareDiagonals(Diag, UDiag, LDiag, a, sx, coef, 1);
			TridiagonalMatrixSolver(Diag, UDiag, LDiag, r, ptr, sx);
		}

		/*************** scan in the y axis ***************/
		size_t shift = 0;
		size_t shift_z = slice - sx;
		const T *d;

		for (size_t z = 0; z < sz; ++z, shift += shift_z)
		{
			for (size_t x = 0; x < sx; ++x, ++shift)
			{
				d = Dd.begin() + shift;
				tmp_ptr = tmp.begin() + shift;
				ptr = out.begin() + shift;

				CopyFrom(tmp_ptr, r, sy, sx);
				PrepareDiagonals(Diag, UDiag, LDiag, d, sy, coef, sx);
				TridiagonalMatrixSolver(Diag, UDiag, LDiag, r, sol, sy);
				Add(sol, ptr, sy, sx);
			}
		}

		/*************** scan in the z axis ***************/
		const T *f;

		for (size_t i = 0; i < slice; ++i)
		{
			f = Df.begin() + i;
			tmp_ptr = tmp.begin() + i;
			ptr = out.begin() + i;

			CopyFrom(tmp_ptr, r, sz, slice);
			PrepareDiagonals(Diag, UDiag, LDiag, f, sz, coef, slice);
			TridiagonalMatrixSolver(Diag, UDiag, LDiag, r, sol, sz);
			AddAndAvg(sol, ptr, sz, slice, avg_coef);
		}

		// memory deallocation
		delete [] r;
		delete [] Diag;
		delete [] UDiag;
		delete [] LDiag;
		delete [] sol;
	}

	/**********************************************************************/

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                           Gauss_LOD                         //
	//                                                             //
	/////////////////////////////////////////////////////////////////

	/**********************************************************************/

	template <class VOXEL> void Gauss_LOD(Image3d<VOXEL> &img, VOXEL sigma, size_t num_iter)
	{
		// the number of iterations must be positive
		if (num_iter == 0)
		{
			throw InternalException("The number of iterations must be a positive integer!");
		}

		// the image size
		size_t sx = img.GetSizeX();
		size_t sy = img.GetSizeY();
		size_t sz = img.GetSizeZ();

		// the image resolution
		i3d::Vector3d<float> res = img.GetResolution().GetRes();
		
		// time step in the x axis
		VOXEL tau_x = VOXEL(0.5) * sigma * sigma / num_iter;

		// accommodate the time step in the y axis with its resolution
		res.y /= res.x;
		VOXEL tau_y = tau_x * res.y * res.y;
	
		// precomputed LR decomposition for the x and y axes
		VOXEL *ptr;
		VOXEL *Mx = new VOXEL[sx];
		VOXEL *Lx = new VOXEL[sx - 1];
		VOXEL *My = new VOXEL[sy];
		VOXEL *Ly = new VOXEL[sy - 1];
		PrepareLRDecomposition(Mx, Lx, sx, tau_x);
		PrepareLRDecomposition(My, Ly, sy, tau_y);

		// 2D images are processed in a different way than 3D ones
		if (sz == 1)
		{
			while (num_iter > 0)
			{
				// filtering in the x axis
				ptr = img.begin();
				for (size_t num = 0; num < sy; ++num, ptr += sx)
				{
					ForwardBackwardSubstitution(Mx, Lx, ptr, sx, tau_x, 1);
				}

				// filtering in the y axis
				ptr = img.begin();
				for (size_t num = 0; num < sx; ++num, ++ptr)
				{
					ForwardBackwardSubstitution(My, Ly, ptr, sy, tau_y, sx);
				}

				--num_iter;
			}
		}
		else
		{
			// the number of voxels in one xy slice
			size_t slice = sx * sy;

			// accommodate the time step in the z axis with its resolution;
			res.z /= res.x;
			VOXEL tau_z = tau_x * res.z * res.z;

			// precomputed LR decomposition for the z axis
			VOXEL *Mz = new VOXEL[sz];
			VOXEL *Lz = new VOXEL[sz - 1];
			PrepareLRDecomposition(Mz, Lz, sz, tau_z);

			while (num_iter > 0)
			{
				// filtering in the x axis
				ptr = img.begin();
				for (size_t num = sy * sz; num > 0; --num, ptr += sx)
				{
					ForwardBackwardSubstitution(Mx, Lx, ptr, sx, tau_x, 1);
				}

				// filtering in the y axis
				ptr = img.begin();
				for (size_t z = 0; z < sz; ++z, ptr += slice - sx)
				{
					for (size_t x = 0; x < sx; ++x, ++ptr)
					{
						ForwardBackwardSubstitution(My, Ly, ptr, sy, tau_y, sx);
					}
				}

				// filtering in the z axis
				ptr = img.begin();
				for (size_t num = slice; num > 0; --num, ++ptr)
				{
					ForwardBackwardSubstitution(Mz, Lz, ptr, sz, tau_z, slice);
				}

				--num_iter;
			}

			// memory deallocation
			delete [] Mz;
			delete [] Lz;
		}

		// memory deallocation
		delete [] Mx;
		delete [] Lx;
		delete [] My;
		delete [] Ly;
	}
	
	/**********************************************************************/

	/** Explicit instantiations. */
	template I3D_ALGO_EXPORT void Gauss_LOD(Image3d<float> &img, float sigma, size_t num_iter);
	template I3D_ALGO_EXPORT void Gauss_LOD(Image3d<double> &img, double sigma, size_t num_iter);

	/**********************************************************************/

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                         GrayGauss_LOD                       //
	//                                                             //
	/////////////////////////////////////////////////////////////////

	/**********************************************************************/

	template <class VOXEL, class PRECISION> void GrayGauss_LOD(Image3d<VOXEL> &img, PRECISION sigma, size_t num_iter)
	{
		Image3d<PRECISION> tmp;
		GrayToFloat(img, tmp);
		Gauss_LOD(tmp, sigma, num_iter);
		FloatToGrayNoWeight(tmp, img);
	}

	/**********************************************************************/

	/** Explicit instantiations. */
	template I3D_ALGO_EXPORT void GrayGauss_LOD(Image3d<GRAY8> &img, float sigma, size_t num_iter);
	template I3D_ALGO_EXPORT void GrayGauss_LOD(Image3d<GRAY16> &img, float sigma, size_t num_iter);
	template I3D_ALGO_EXPORT void GrayGauss_LOD(Image3d<GRAY8> &img, double sigma, size_t num_iter);
	template I3D_ALGO_EXPORT void GrayGauss_LOD(Image3d<GRAY16> &img, double sigma, size_t num_iter);

	/**********************************************************************/

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                             EED_AOS                         //
	//                                                             //
	/////////////////////////////////////////////////////////////////

	/**********************************************************************/

	template <class VOXEL> void EED_AOS(Image3d<VOXEL> &img, VOXEL sigma, VOXEL lambda, VOXEL tau, size_t num_iter)
	{
		// 3D images are processed in a different way than 2D ones
		if (img.GetSizeZ() > 1)
		{
			// prepare temporary images
			i3d::Image3d<VOXEL> tmp, Da, Db, Dc, Dd, De, Df;
			Da.CopyMetaData(img);
			Db.CopyMetaData(img);
			Dc.CopyMetaData(img);
			Dd.CopyMetaData(img);
			De.CopyMetaData(img);
			Df.CopyMetaData(img);

			// execute given number of iterations
			while (num_iter > 0)
			{
				// store the input image for future reference
				tmp = img;

				// apply the Gaussian filter before the gradient estimation
				GaussIIR(img, sigma);

				// construct a structure tensor
				StructureTensor3D(img, Da, Db, Dc, Dd, De, Df);

				// construct a diffusion tensor
				EEDTensor3D(Da, Db, Dc, Dd, De, Df, lambda);

				// execute one step of the anisotropic diffusion based on the semi-implicit AOS-stabilized scheme
				AnisoDiffStep3D_AOS(tmp, Da, Db, Dc, Dd, De, Df, img, tau);

				--num_iter;
			}
		}
		else
		{
			// prepare temporary images
			i3d::Image3d<VOXEL> tmp, Da, Db, Dc;
			Da.CopyMetaData(img);
			Db.CopyMetaData(img);
			Dc.CopyMetaData(img);

			// execute given number of iterations
			while (num_iter > 0)
			{
				// store the input image for future reference
				tmp = img;

				// apply the Gaussian filter before the gradient estimation
				GaussIIR(img, sigma);

				// construct a structure tensor
				StructureTensor(img, Da, Db, Dc);

				// construct a diffusion tensor
				EEDTensor(Da, Db, Dc, lambda);

				// execute one step of the anisotropic diffusion based on the semi-implicit AOS-stabilized scheme
				AnisoDiffStep_AOS(tmp, Da, Db, Dc, img, tau);

				--num_iter;
			}
		}
	}
	
	/**********************************************************************/

	/** Explicit instantiations. */
	template I3D_ALGO_EXPORT void EED_AOS(Image3d<float> &img, float sigma, float lambda, float tau, size_t num_iter);
	template I3D_ALGO_EXPORT void EED_AOS(Image3d<double> &img, double sigma, double lambda, double tau, size_t num_iter);

	/**********************************************************************/

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                          GrayEED_AOS                        //
	//                                                             //
	/////////////////////////////////////////////////////////////////

	/**********************************************************************/

	template <class VOXEL, class PRECISION> void GrayEED_AOS(Image3d<VOXEL> &img, PRECISION sigma, PRECISION lambda, PRECISION tau, size_t num_iter)
	{
		Image3d<PRECISION> tmp;
		GrayToFloat(img, tmp);
		EED_AOS(tmp, sigma, lambda, tau, num_iter);
		FloatToGrayNoWeight(tmp, img);
	}

	/**********************************************************************/

	/** Explicit instantiations. */
	template I3D_ALGO_EXPORT void GrayEED_AOS(Image3d<GRAY8> &img, float sigma, float lambda, float tau, size_t num_iter);
	template I3D_ALGO_EXPORT void GrayEED_AOS(Image3d<GRAY16> &img, float sigma, float lambda, float tau, size_t num_iter);
	template I3D_ALGO_EXPORT void GrayEED_AOS(Image3d<GRAY8> &img, double sigma, double lambda, double tau, size_t num_iter);
	template I3D_ALGO_EXPORT void GrayEED_AOS(Image3d<GRAY16> &img, double sigma, double lambda, double tau, size_t num_iter);

	/**********************************************************************/

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                             CED_AOS                         //
	//                                                             //
	/////////////////////////////////////////////////////////////////

	/**********************************************************************/

	template <class VOXEL> void CED_AOS(Image3d<VOXEL> &img, VOXEL sigma, VOXEL rho, VOXEL tau, size_t num_iter)
	{
		// 3D images are processed in a different way than 2D ones
		if (img.GetSizeZ() > 1)
		{
			// prepare temporary images
			i3d::Image3d<VOXEL> tmp, Da, Db, Dc, Dd, De, Df;
			Da.CopyMetaData(img);
			Db.CopyMetaData(img);
			Dc.CopyMetaData(img);
			Dd.CopyMetaData(img);
			De.CopyMetaData(img);
			Df.CopyMetaData(img);

			// execute given number of iterations
			while (num_iter > 0)
			{
				// store the input image for future reference
				tmp = img;

				// apply the Gaussian filter before the gradient estimation
				GaussIIR(img, sigma);
				
				// construct structure tensors
				StructureTensor3D(img, Da, Db, Dc, Dd, De, Df);

				// apply the Gaussian filter in order to smooth the structure tensors
				GaussIIR(Da, rho);
				GaussIIR(Db, rho);
				GaussIIR(Dc, rho);
				GaussIIR(Dd, rho);
				GaussIIR(De, rho);
				GaussIIR(Df, rho);

				// construct diffusion tensors
				CEDTensor3D(Da, Db, Dc, Dd, De, Df);

				// execute one step of the anisotropic diffusion based on the semi-implicit AOS-stabilized scheme
				AnisoDiffStep3D_AOS(tmp, Da, Db, Dc, Dd, De, Df, img, tau);

				--num_iter;
			}
		}
		else
		{
			// prepare temporary images
			i3d::Image3d<VOXEL> tmp, Da, Db, Dc;
			Da.CopyMetaData(img);
			Db.CopyMetaData(img);
			Dc.CopyMetaData(img);

			// execute given number of iterations
			while (num_iter > 0)
			{
				// store the input image for future reference
				tmp = img;

				// apply the Gaussian filter before the gradient estimation
				GaussIIR(img, sigma);

				// construct structure tensors
				StructureTensor(img, Da, Db, Dc);
				
				// apply the Gaussian filter in order to smooth the structure tensors
				GaussIIR(Da, rho);
				GaussIIR(Db, rho);
				GaussIIR(Dc, rho);
				
				// construct diffusion tensors
				CEDTensor(Da, Db, Dc);

				// execute one step of the anisotropic diffusion based on the semi-implicit AOS-stabilized scheme
				AnisoDiffStep_AOS(tmp, Da, Db, Dc, img, tau);

				--num_iter;
			}
		}
	}
	
	/**********************************************************************/

	/** Explicit instantiations. */
	template I3D_ALGO_EXPORT void CED_AOS(Image3d<float> &img, float sigma, float rho, float tau, size_t num_iter);
	template I3D_ALGO_EXPORT void CED_AOS(Image3d<double> &img, double sigma, double rho, double tau, size_t num_iter);

	/**********************************************************************/

	/////////////////////////////////////////////////////////////////
	//                                                             //
	//                          GrayCED_AOS                        //
	//                                                             //
	/////////////////////////////////////////////////////////////////

	/**********************************************************************/

	template <class VOXEL, class PRECISION> void GrayCED_AOS(Image3d<VOXEL> &img, PRECISION sigma, PRECISION rho, PRECISION tau, size_t num_iter)
	{
		Image3d<PRECISION> tmp;
		GrayToFloat(img, tmp);
		CED_AOS(tmp, sigma, rho, tau, num_iter);
		FloatToGrayNoWeight(tmp, img);
	}

	/**********************************************************************/

	/** Explicit instantiations. */
	template I3D_ALGO_EXPORT void GrayCED_AOS(Image3d<GRAY8> &img, float sigma, float rho, float tau, size_t num_iter);
	template I3D_ALGO_EXPORT void GrayCED_AOS(Image3d<GRAY16> &img, float sigma, float rho, float tau, size_t num_iter);
	template I3D_ALGO_EXPORT void GrayCED_AOS(Image3d<GRAY8> &img, double sigma, double rho, double tau, size_t num_iter);
	template I3D_ALGO_EXPORT void GrayCED_AOS(Image3d<GRAY16> &img, double sigma, double rho, double tau, size_t num_iter);

	/**********************************************************************/
	
} // namespace i3d





