/*
* i3dlib - image manipulation library
*
* Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Library General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Library General Public License for more details.
*
* You should have received a copy of the GNU Library General Public
* License along with this library; if not, write to the Free
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
* FILE: rankFilterByHistogram.cc
*
* Image filtering using rank filters
*
* Radoslav Stefanik <256711@mail.muni.cz> 2010
*/


#include <limits.h>
#include <vector>
#include <map>

#include "image3d.h"
#include "rankFilterByHistogram.h"
#include "neighbours.h"

namespace i3d {

    /***************************************************************************/
    /**
    this class iterates through all the points in one plane given by normal to this plane
    */
    class FieldIter {
private:
        int i,j;
        Vector3d<int> dirI, dirJ;
        int maxI, maxJ;
public:
        FieldIter(const Vector3d<int>& direction, const Vector3d<int>& bounds);
        int getNextPosition(Vector3d<int>& nextPosition);
    };

    /***************************************************************************/
    /**
    constructor gets one point that is the normal to the desired pane, and point maximum determinating limits of the plane
    */
    FieldIter::FieldIter(const Vector3d<int>& direction, const Vector3d<int>& bounds) {
        i = j = 0;
        //sets two vectors perpendicular to direction and their bounds in image
        Vector3d<int> dirs[] = { Vector3d<int>(0,0,1), Vector3d<int>(0,1,0), Vector3d<int>(1,0,0) };
        if (direction.x > 0) {
            dirI = dirs[0], maxI = bounds.z;
            dirJ = dirs[1], maxJ = bounds.y;
        } else if (direction.y>0) {
            dirI = dirs[0], maxI = bounds.z;
            dirJ = dirs[2], maxJ = bounds.x;
        } else {
            dirI = dirs[1], maxI = bounds.y;
            dirJ = dirs[2], maxJ = bounds.x;
        }
    }

    /***************************************************************************/
    /**
    assign next point to p, if there is no more points in the plane it returns 1 otherwise returns 0
    */
    int FieldIter::getNextPosition(Vector3d<int>& nextPosition) {
        if (j >= maxJ) return 1;
        nextPosition = i * dirI + j * dirJ;
        i++;
        if (i >= maxI) {
            i=0;
            j++;
        }
        return 0;
    }

    /***************************************************************************/
    /**
    function that analyzes shape of SE
    decides which direction is best for moving
    and computes points we will add and remove in each shift
    */
    void analyze(const std::vector<Vector3d<int> >& SE, std::vector< Vector3d<int> >& oldPoints,
                 std::vector<Vector3d<int> >& newPoints, Vector3d<int>& direction, Vector3d<int>& SEmin, Vector3d<int>& SEmax) {
        SEmin.x = SEmin.y = SEmin.z = INT_MAX;
        SEmax.x = SEmax.y = SEmax.z = INT_MIN;
        //find the bounds of SE
        for (int i=0; i<(int)SE.size(); i++) {
            if (SE[i].x>SEmax.x) SEmax.x=SE[i].x;
            if (SE[i].x<SEmin.x) SEmin.x=SE[i].x;
            if (SE[i].y>SEmax.y) SEmax.y=SE[i].y;
            if (SE[i].y<SEmin.y) SEmin.y=SE[i].y;
            if (SE[i].z>SEmax.z) SEmax.z=SE[i].z;
            if (SE[i].z<SEmin.z) SEmin.z=SE[i].z;
        }
        //create bounding box of SE as 3D array that represents SE
        Vector3d<int> bounds (SEmax.x-SEmin.x+1, SEmax.y-SEmin.y+1, SEmax.z-SEmin.z+1);
        bool*** mySE = new bool**[bounds.x];
        for (int i=0; i<bounds.x; i++) {
            mySE[i] = new bool*[bounds.y];
            for (int j=0; j<bounds.y; j++) {
                mySE[i][j] = new bool[bounds.z];
                for (int k=0; k<bounds.z; k++) mySE[i][j][k]=0;
            }
        }
        //points that belong to SE are represented as 1
        for (size_t i=0; i<SE.size(); i++) {
            mySE[SEmax.x - SE[i].x][SEmax.y - SE[i].y][SEmax.z - SE[i].z] = 1;
        }
        Vector3d<int> dirs[] = { Vector3d<int>(1,0,0), Vector3d<int>(0,1,0), Vector3d<int>(0,0,1) };
        int minshifts = INT_MAX;
        //int mindirection = -1;

        //we check how many elements is added or removed in one shift for each direction
        for (int i=0; i<3; i++) {
            FieldIter iter(dirs[i],bounds);
            Vector3d<int> p(0,0,0);
            int shifts=0;
            std::vector< Vector3d<int> > oldTmp;
            std::vector< Vector3d<int> > newTmp;
            Vector3d<int> tmp;
            while (!iter.getNextPosition(p)) {
                bool prev=0;
                int m;
                for (m=0; m<Dot(bounds,dirs[i]); m++) {
                    tmp = p + m * dirs[i];
                    bool next = mySE[tmp.x][tmp.y][tmp.z];
                    //we set original coordinates for saved point
                    tmp.x += SEmin.x;
                    tmp.y += SEmin.y;
                    tmp.z += SEmin.z;
                    if (!prev && next) {
                        shifts++;
                        oldTmp.push_back(tmp);
                    } else if (prev && !next) {
                        newTmp.push_back(tmp);
                    }
                    prev=next;
                }
                //if pixel is on the edge of SE we didnt count it in new points so we add it now
                if (prev) {
                    tmp = p + m * dirs[i];
                    tmp.x += SEmin.x;
                    tmp.y += SEmin.y;
                    tmp.z += SEmin.z;
                    newTmp.push_back(tmp);
                }
            }
            if (shifts < minshifts) {
                minshifts = shifts;
                //mindirection = i;
                direction = dirs[i];
                oldPoints = oldTmp;
                newPoints = newTmp;
            }
        }
        //we check newPoints for new bounds of SE that will be used in calculation
        for (unsigned int i=0; i<newPoints.size(); i++) {
            if (newPoints[i].x > SEmax.x) SEmax.x = newPoints[i].x;
            if (newPoints[i].y > SEmax.y) SEmax.y = newPoints[i].y;
            if (newPoints[i].z > SEmax.z) SEmax.z = newPoints[i].z;
        }


        //deleting allocated array
        for (int i=0; i<bounds.x; i++) {
            for (int j=0; j<bounds.y; j++) {
                delete[] mySE[i][j];
            }
            delete[] mySE[i];
        }
        delete[] mySE;
    }

    /***************************************************************************/
    /**
    Applies rank filter on the input image using given neighbourhood.
    Result is stored in the output image (this image is going to be resized according to the input).
    */
    template<typename T>
    void rankFilterByHistogram(const i3d::Image3d<T>& imgIn, i3d::Image3d<T>& imgOut,
            Neighbourhood& SEin, int rank) {

        std::vector< Vector3d<int> > SE = SEin.offset;
        if (rank>(int)SE.size()||rank<0) {
            std::cerr<< "Error: rank must be from <1,|SE|>" << std::endl;
            if (rank > (int)SE.size()) rank=SE.size();
            return;
        }
        //resize output image
        imgOut.MakeRoom(imgIn.GetSizeX(), imgIn.GetSizeY(), imgIn.GetSizeZ());
        imgOut.SetOffset(imgIn.GetOffset());
        imgOut.SetResolution(imgIn.GetResolution());

        Vector3d<int> dir, SEmin, SEmax;
        std::vector<Vector3d<int> > oldPositions;
        std::vector<Vector3d<int> > newPositions;


        //analyzes SE, chooses ideal direction of movement and computes positions that will be added or removed each step
        analyze(SE, oldPositions, newPositions, dir, SEmin, SEmax);

        Vector3d<int> bounds;
        bounds.x = imgIn.GetWidth();
        bounds.y = imgIn.GetHeight();
        bounds.z = imgIn.GetNumSlices();

        FieldIter iter(dir,bounds);
        int boundI = Dot(dir, bounds);

        int numOfElements;
        std::map<T, int> histogram;
        Vector3d<int> basePoint;

        //we take first point on the side of picture and process all the points that are covered by SE
        //we are moving in ideal direction we calclated
        while (iter.getNextPosition(basePoint)==0) {
            histogram.clear();
            numOfElements =0 ;
            //adding all elements in SE for first pixel we are iterating
            for (size_t i=0; i<SE.size(); i++) {
                Vector3d<int> tmp = basePoint + SE[i];
                if (tmp.x<0 || tmp.x>=bounds.x || tmp.y<0 || tmp.y>=bounds.y || tmp.z<0 || tmp.z>=bounds.z) continue;
                histogram[imgIn.GetVoxel(tmp.x, tmp.y, tmp.z)]++;
                ++numOfElements;
            }
            T currValue=0;
            int beforePtr;
            int actRank = rank;
            if (numOfElements != (int)SE.size()) {
                //compute new rank if there are not all points of SE inside bounds of image
                actRank = static_cast<int>(std::floor (((double)numOfElements/(double)SE.size()) * rank +0.5));
                if( actRank<=0) actRank=1;
            }

            typename std::map<T, int>::iterator iter;
            typename std::map<T, int>::iterator ptrToRankElement;

            //initialization of pointers in histogram
            //if there are no points of SE inside the image output is 0
            if (numOfElements <= 0) {
                currValue=0;
                beforePtr=0;
                ptrToRankElement = iter = histogram.begin();
            //choose ideal direction for passing histogram
            } else if (actRank <= numOfElements/2 + 1) {
                //we iterate to the rankth element in histogram from begining
                ptrToRankElement = iter = histogram.begin();
                int elements;
                for ( elements=0; elements < actRank && iter!=histogram.end(); ) {
                    ptrToRankElement = iter;
                    elements+=iter->second;
                    if (iter == histogram.end()) break;
                    iter++;
                }
                currValue = ptrToRankElement->first;
                beforePtr = elements - ptrToRankElement->second;
            } else {
                //passing through histogram to rankth element from end
                ptrToRankElement = histogram.end();
                beforePtr = numOfElements - ptrToRankElement->second;
                while(beforePtr >= actRank) {
                    ptrToRankElement--;
                    beforePtr -= ptrToRankElement->second;
                }
                currValue = ptrToRankElement->first;
            }
            imgOut.SetVoxel(basePoint.x, basePoint.y, basePoint.z, currValue);

            //histogram on edge of picture is loaded,we can iterate through other pixels using our computed points
            for (int i=0; i<boundI-1; i++) {
                Vector3d<int> actualPoint = basePoint + i * dir;
                //we check bounding box of SE if it is out of image and when it is we check for each point
                //if whole SE is inside the image we do not check every point of it for bounds
                bool outOfBounds=false;
                Vector3d<int> tmp = actualPoint + SEmin;
                if (tmp.x<0 || tmp.x>=bounds.x || tmp.y<0 || tmp.y>=bounds.y || tmp.z<0 || tmp.z>=bounds.z) outOfBounds=true;
                tmp = actualPoint + SEmax;
                if (tmp.x<0 || tmp.x>=bounds.x || tmp.y<0 || tmp.y>=bounds.y || tmp.z<0 || tmp.z>=bounds.z) outOfBounds=true;

                //adding elements to histogram
                for (size_t j=0; j<newPositions.size(); j++) {
                    tmp = actualPoint + newPositions[j];
                    if (outOfBounds && (tmp.x<0 || tmp.x>=bounds.x || tmp.y<0 || tmp.y>=bounds.y || tmp.z<0 || tmp.z>=bounds.z) ) continue;
                    T newPoint = imgIn.GetVoxel(tmp.x, tmp.y, tmp.z);
                    histogram[newPoint]++;
                    numOfElements++;
                    if (newPoint< currValue) beforePtr++;
                }

                //removing elements from histogram
                for (size_t j=0; j<oldPositions.size(); j++) {
                    tmp = actualPoint + oldPositions[j];
                    if (outOfBounds && (tmp.x<0 || tmp.x>=bounds.x || tmp.y<0 || tmp.y>=bounds.y || tmp.z<0 || tmp.z>=bounds.z) ) continue;
                    T oldPoint = imgIn.GetVoxel(tmp.x, tmp.y, tmp.z);
                    //we insert point in our histogram
                    if (oldPoint < currValue) beforePtr--;
                    numOfElements--;
                    //if we delete element we are pointing to, we must move the pointer to next element
                    if (--histogram[oldPoint] <= 0) {
                        if (ptrToRankElement->first == oldPoint ) {
                            if (ptrToRankElement != histogram.end()) {
                                ptrToRankElement++;
                                histogram.erase(oldPoint);
                            } else {
                                //if we delete last element in histogram and we have pointer set to it we must set i to new end
                                histogram.erase(oldPoint);
                                ptrToRankElement = histogram.end();
                                beforePtr-=ptrToRankElement->second;
                            }
                        } else histogram.erase(oldPoint);
                    }
                }

                //actualize number of items in histogram
                actRank = rank;
                //if we are on edge we modify the rank
                if (numOfElements != (int)SE.size()) {
                    actRank = static_cast<int>(std::floor (((double)numOfElements/(double)SE.size()) * rank +0.5));
                    if(actRank<=0) actRank=1;
                }
                //if there are no pixels in histogram we set value to 0
                if (numOfElements<=0) {
                    currValue=0;
                    beforePtr=0;
                    ptrToRankElement = iter = histogram.begin();
                } else {
                    //if pointer is too much in forward we must increase it
                    int numOfVals = ptrToRankElement->second;
                    if ( beforePtr + numOfVals < actRank) {
                        while( beforePtr+ ptrToRankElement->second < actRank) {
                            beforePtr += ptrToRankElement->second;
                            ptrToRankElement++;
                        }
                        //if pointer is in back we must decrease it
                    } else if (beforePtr >= actRank) {
                        while(beforePtr >= actRank) {
                            ptrToRankElement--;
                            beforePtr -= ptrToRankElement->second;
                        }
                    }
                }

                currValue = ptrToRankElement->first;
                //increase position of processed pixel in direction we are moving
                actualPoint = basePoint + (i + 1) * dir;
                imgOut.SetVoxel(actualPoint.x, actualPoint.y, actualPoint.z, currValue);
            }
        }
        //std::cerr<<"RankFilter done" << std::endl;
    }

    /***************************************************************************/

    /* Explicit instantiations: */



    template I3D_ALGO_EXPORT void rankFilterByHistogram(const i3d::Image3d<i3d::GRAY8>& imgIn, i3d::Image3d<i3d::GRAY8>& imgOut, Neighbourhood& SEin, int rank);
    template I3D_ALGO_EXPORT void rankFilterByHistogram(const i3d::Image3d<i3d::GRAY16>& imgIn, i3d::Image3d<i3d::GRAY16>& imgOut, Neighbourhood& SEin, int rank);


}
