/*
 * i3dlib - image manipulation library
 *
 * Copyright (C) 2000-2006   Centre for Biomedical Image Analysis (CBIA)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file watershed_beucher.cc
  \brief Part of mathematical morphology routines - implementation.

  This file contains bodies of mathematical morphology routines dealing with 
  the watershed transformation algorithm by Beucher.
*/ 

#include <vector>
#include <queue>
#include <map>
#include <limits>

#include "watersheds.h"
#include "watersheds_internal.h"
#include "i3dassert.h"

#ifdef __GNUG__
#	pragma implementation
#endif

namespace i3d {

	//--------------------------------------------------------------------------
	// Watershed Beucher's algorithm implementation
	//--------------------------------------------------------------------------

	/** Priority queue used in the Beucher's watershed algorithm. This is a 
	hierarchical queue with a separator (special "alpha" queue) dividing 
	different generations. For detailed function description, see the article
	cited in the comment of the WatershedBeucher function.
	*/
	template <typename T>
	class CBeucherQueue
	{
	public:
		/// creates new empty queue - allocates some structures and prepares for processing
		CBeucherQueue()
		{
			alphaQueue = new std::queue<size_t>();
		}
		/// frees all the allocated resources
		~CBeucherQueue()
		{
			delete alphaQueue;
		}

		/// pushes voxel with given priority into the queue
		void Push(const T &priority, const size_t &value)
		{
			typename std::map<T, std::queue<size_t>* >::iterator priorityPair = priorityQueue.find(priority);
			if (priorityPair == priorityQueue.end())
			{
				std::queue<size_t> *newQueue = new std::queue<size_t>();
				newQueue->push(value);
				priorityQueue[priority] = newQueue;
			}
			else
				priorityPair->second->push(value);
		}
		/// pushes voxel into the alpha queue
		void PushAlpha(const size_t &value)
		{
			alphaQueue->push(value);
		}
		/// returns and pops the first element with the highest priority (the lowest priority value) from the queue
		size_t TopAndPop()
		{
			I3DASSERT_BT(!IsEmpty(), "You are trying to top and pop an element from an empty queue.");
			std::queue<size_t>* firstQueue = priorityQueue.begin()->second;
			size_t result = firstQueue->front();
			firstQueue->pop();
			return result;
		}
		/// returns whether is highest priority queue empty or not
		bool IsHighestPriorityEmpty() const
		{
			return priorityQueue.begin()->second->empty();
		}
		/// returns whether is alpha queue empty or not
		bool IsAlphaEmpty() const
		{
			return alphaQueue->empty();
		}
		/// returns whether is this queue empty or not
		bool IsEmpty() const
		{
			return priorityQueue.empty();
		}
		/// returns the highest (the most prior :) stored priority (the lowest value)
		T GetHighestPriority() const
		{
			I3DASSERT_BT(!IsEmpty(), "You are trying to get the lowest priority from an empty queue.");
			return priorityQueue.begin()->first;
		}
		/// switches alpha queue and the highest priority queue
		void SwitchAlphaAndHighestPriority()
		{
			I3DASSERT_BT(!IsEmpty(), "You are trying to switch the alpha queue in an empty queue.");
			std::queue<size_t>* firstQueue = priorityQueue.begin()->second;
			priorityQueue.begin()->second = alphaQueue;
			alphaQueue = firstQueue;
		}
		/// removes the highest priority queue
		void RemoveHighestPriority()
		{
			I3DASSERT_BT(!IsEmpty(), "You are trying to remove highest priority queue from an empty queue.");
			typename std::map<T, std::queue<size_t>* >::iterator firstPair = priorityQueue.begin();
			delete firstPair->second;
			priorityQueue.erase(firstPair);
		}

	private:
		/// alpha queue (separator of the generations), always has the second highest priority
		std::queue<size_t> *alphaQueue;
		/// priority queue (for normal opperation)
		std::map<T, std::queue<size_t>* > priorityQueue;
	};

	template <typename IN_TYPE, typename OUT_TYPE> 
	I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<IN_TYPE> &input, const Image3d<OUT_TYPE> &marker, Image3d<OUT_TYPE> &output, const Neighbourhood &neighbourhood, bool drawWsLines)
	{
		// mapping of the variable names in the paper to this implementation names:
		// f = inputCopy
		// g = output
		// C = collisionLabel
		// k = currentLevel
		// n = generation

		I3DASSERT_ALWAYS_ET(input.GetImageSize() != 0, "Input image should not be empty.");
		I3DASSERT_ALWAYS_ET(input.GetImageSize() == marker.GetImageSize(), "Input and marker image size differs.");

		// this algorithm works on the input copy - we want to be able to write into the image, during the copying, get the highest voxel value
		Image3d<IN_TYPE> inputCopy;
		inputCopy.CopyMetaData(input);
		size_t imageSize = input.GetImageSize();
		IN_TYPE maxVoxelValue = (!std::numeric_limits<IN_TYPE>::is_integer && std::numeric_limits<IN_TYPE>::has_denorm == std::denorm_present ? IN_TYPE(0) - std::numeric_limits<IN_TYPE>::max() : std::numeric_limits<IN_TYPE>::min());
		for (size_t i = 0; i < imageSize; ++i)
		{
			IN_TYPE value = input.GetVoxel(i);
			maxVoxelValue = std::max(maxVoxelValue, value);
			inputCopy.SetVoxel(i, value);
		}
		// copy marker into output (correct labels to the form we need - odd markers), determine the collision label
		output.CopyMetaData(input);
		OUT_TYPE collisionLabel = 0;
		for (size_t i = 0; i < imageSize; ++i)
		{
			OUT_TYPE value = marker.GetVoxel(i);
			collisionLabel = std::max(collisionLabel, value);
			if (value)
				output.SetVoxel(i, value * 2 - 1);
			else
				output.SetVoxel(i, 0);
		}
		I3DASSERT_ALWAYS_ET(2 * collisionLabel + 1 > collisionLabel, "Labels overflow, marker image contains too many (or too high) labels.");
		collisionLabel = 2 * collisionLabel + 1;
		// initialize the hierarchy queue - insert labeled voxels with some neighbour different than they are, voxels with all the same neighbors are marked as final
		CBeucherQueue<IN_TYPE> beucherQueue;
		NeighbourhoodWalker nbhWalker(input.GetSize(), neighbourhood, drawWsLines);
		for (size_t i = 0; i < imageSize; ++i)
		{
			OUT_TYPE label;
			if (!(label = output.GetVoxel(i)))
				continue;
			nbhWalker.Begin(i);
			while (nbhWalker.Valid())
			{
				if (output.GetVoxel(nbhWalker.GetIndex()) != label)
				{
					label = 0;
					break;
				}
				nbhWalker.Next();
			}
			if (label)
				output.SetVoxel(i, label + 1);
			else
				beucherQueue.Push(inputCopy.GetVoxel(i), i);
		}
		// process all the voxels that needs to be (and can be) processed
		while (!beucherQueue.IsEmpty())
		{
			IN_TYPE currentLevel = beucherQueue.GetHighestPriority();
			IN_TYPE generation = IN_TYPE(1);
			// try to precess all the voxels from the current level and the alpha queue
			while (!beucherQueue.IsHighestPriorityEmpty())
			{
				// try to precess all the voxels from the current level
				while (!beucherQueue.IsHighestPriorityEmpty())
				{
					// process the first highest priority voxel
					size_t i = beucherQueue.TopAndPop();
					OUT_TYPE label = output.GetVoxel(i);
					if (label != collisionLabel)
					{
						label += 1;
						output.SetVoxel(i, label);
						nbhWalker.Begin(i);
						while (nbhWalker.Valid())
						{
							size_t j = nbhWalker.GetIndex();
							OUT_TYPE neighbourLabel = output.GetVoxel(j);
							if (neighbourLabel == 0)
							{
								if (inputCopy.GetVoxel(j) > currentLevel)
									beucherQueue.Push(inputCopy.GetVoxel(j), j);
								else
								{
									beucherQueue.PushAlpha(j);
									I3DASSERT_ALWAYS_ET(std::numeric_limits<IN_TYPE>::has_infinity ? ((maxVoxelValue + generation) != std::numeric_limits<IN_TYPE>::infinity()) : ((maxVoxelValue + generation) > maxVoxelValue), "Image value overflow, input image contains higher values than it should.");
									inputCopy.SetVoxel(j, maxVoxelValue + generation);
								}
								output.SetVoxel(j, label - 1);
							}
							else
							{
								if ((neighbourLabel % 2 == 1) && (inputCopy.GetVoxel(i) != inputCopy.GetVoxel(j)) && (neighbourLabel != label - 1))
									output.SetVoxel(j, collisionLabel);
							}
							nbhWalker.Next();
						}
						if (generation > IN_TYPE(1))
							inputCopy.SetVoxel(i, currentLevel);
					}
					else
						inputCopy.SetVoxel(i, currentLevel);
				}
				if (!beucherQueue.IsAlphaEmpty())
				{
					beucherQueue.SwitchAlphaAndHighestPriority();
					generation += IN_TYPE(1);
				}
			}
			// both highest priority and the alpha queue are empty, so we can remove the highest priority queue
			beucherQueue.RemoveHighestPriority();
		}
		// convert labels in the output image back, replace collision label with the watershed line label
		for (size_t i = 0; i < imageSize; ++i)
		{
			OUT_TYPE value = output.GetVoxel(i);
			if (value != collisionLabel)
				output.SetVoxel(i, (value + 1 ) / 2);
			else
				output.SetVoxel(i, 0);
		}
		// draw the watershed lines if requested to
		if (drawWsLines)
		{
			std::vector<size_t> wsLine;
			for (size_t i = 0; i < imageSize; ++i)
			{
				OUT_TYPE label = output.GetVoxel(i);
				if (label)
				{
					// for voxels with some label and some neighbours of the different label, do the 
					nbhWalker.Begin(i);
					while (nbhWalker.Valid())
					{
						OUT_TYPE nbhLabel = output.GetVoxel(nbhWalker.GetIndex());
						if (nbhLabel != label && nbhLabel != 0)
						{
							wsLine.push_back(i);
							break;
						}
						nbhWalker.Next();
					}
				}
			}
			size_t wsLineSize = wsLine.size();
			for (std::vector<size_t>::size_type i = 0; i < wsLineSize; ++i)
			{
				output.SetVoxel(wsLine[i], 0);
			}
		}
	}

	template <typename T> 
	I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<T> &input, Image3d<size_t> &output, const Neighbourhood &neighbourhood, bool drawWsLines)
	{
		// detect regional minima and construct the marker image
		Image3d<size_t> marker;
		WSInternalMarkerFromMinima(input, marker, neighbourhood);

		// call the watershed
		WatershedBeucher(input, marker, output, neighbourhood, drawWsLines);
	}

	template I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<i3d::GRAY8> &input, const Image3d<size_t> &marker, Image3d<size_t> &output, const Neighbourhood &neighbourhood, bool drawWsLines);
	template I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<i3d::GRAY16> &input, const Image3d<size_t> &marker, Image3d<size_t> &output, const Neighbourhood &neighbourhood, bool drawWsLines);
	template I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<float> &input, const Image3d<size_t> &marker, Image3d<size_t> &output, const Neighbourhood &neighbourhood, bool drawWsLines);
	template I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<i3d::GRAY8> &input, const Image3d<GRAY16> &marker, Image3d<GRAY16> &output, const Neighbourhood &neighbourhood, bool drawWsLines);
	template I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<i3d::GRAY16> &input, const Image3d<GRAY16> &marker, Image3d<GRAY16> &output, const Neighbourhood &neighbourhood, bool drawWsLines);
	template I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<float> &input, const Image3d<GRAY16> &marker, Image3d<GRAY16> &output, const Neighbourhood &neighbourhood, bool drawWsLines);
	template I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<i3d::GRAY8> &input, Image3d<size_t> &output, const Neighbourhood &neighbourhood, bool drawWsLines);
	template I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<i3d::GRAY16> &input, Image3d<size_t> &output, const Neighbourhood &neighbourhood, bool drawWsLines);
	template I3D_ALGO_EXPORT void WatershedBeucher(const Image3d<float> &input, Image3d<size_t> &output, const Neighbourhood &neighbourhood, bool drawWsLines);

} // namespace i3d
